"""
Support for RallyDev

.. versionadded:: 2015.8.0

Requires a ``username`` and a ``password`` in ``/etc/salt/minion``:

.. code-block:: yaml

    rallydev:
      username: myuser@example.com
      password: 123pass
"""

import logging

import salt.utils.http
import salt.utils.json
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if apache is installed
    """
    if not __opts__.get("rallydev", {}).get("username", None):
        return (
            False,
            "The rallydev execution module failed to load: rallydev:username not"
            " defined in config.",
        )
    if not __opts__.get("rallydev", {}).get("password", None):
        return (
            False,
            "The rallydev execution module failed to load: rallydev:password not"
            " defined in config.",
        )
    return True


def _get_token():
    """
    Get an auth token
    """
    username = __opts__.get("rallydev", {}).get("username", None)
    password = __opts__.get("rallydev", {}).get("password", None)
    path = "https://rally1.rallydev.com/slm/webservice/v2.0/security/authorize"
    result = salt.utils.http.query(
        path,
        decode=True,
        decode_type="json",
        text=True,
        status=True,
        username=username,
        password=password,
        cookies=True,
        persist_session=True,
        opts=__opts__,
    )
    if "dict" not in result:
        return None

    return result["dict"]["OperationResult"]["SecurityToken"]


def _query(
    action=None, command=None, args=None, method="GET", header_dict=None, data=None
):
    """
    Make a web call to RallyDev.
    """
    token = _get_token()
    username = __opts__.get("rallydev", {}).get("username", None)
    password = __opts__.get("rallydev", {}).get("password", None)
    path = "https://rally1.rallydev.com/slm/webservice/v2.0/"

    if action:
        path += action

    if command:
        path += f"/{command}"

    log.debug("RallyDev URL: %s", path)

    if not isinstance(args, dict):
        args = {}

    args["key"] = token

    if header_dict is None:
        header_dict = {"Content-type": "application/json"}

    if method != "POST":
        header_dict["Accept"] = "application/json"

    decode = True
    if method == "DELETE":
        decode = False

    return_content = None
    result = salt.utils.http.query(
        path,
        method,
        params=args,
        data=data,
        header_dict=header_dict,
        decode=decode,
        decode_type="json",
        text=True,
        status=True,
        username=username,
        password=password,
        cookies=True,
        persist_session=True,
        opts=__opts__,
    )
    log.debug("RallyDev Response Status Code: %s", result["status"])
    if "error" in result:
        log.error(result["error"])
        return [result["status"], result["error"]]

    return [result["status"], result.get("dict", {})]


def list_items(name):
    """
    List items of a particular type

    CLI Examples:

    .. code-block:: bash

        salt myminion rallydev.list_<item name>s
        salt myminion rallydev.list_users
        salt myminion rallydev.list_artifacts
    """
    status, result = _query(action=name)
    return result


def query_item(name, query_string, order="Rank"):
    """
    Query a type of record for one or more items. Requires a valid query string.
    See https://rally1.rallydev.com/slm/doc/webservice/introduction.jsp for
    information on query syntax.

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.query_<item name> <query string> [<order>]
        salt myminion rallydev.query_task '(Name contains github)'
        salt myminion rallydev.query_task '(Name contains reactor)' Rank
    """
    status, result = _query(action=name, args={"query": query_string, "order": order})
    return result


def show_item(name, id_):
    """
    Show an item

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.show_<item name> <item id>
    """
    status, result = _query(action=name, command=id_)
    return result


def update_item(name, id_, field=None, value=None, postdata=None):
    """
    Update an item. Either a field and a value, or a chunk of POST data, may be
    used, but not both.

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.update_<item name> <item id> field=<field> value=<value>
        salt myminion rallydev.update_<item name> <item id> postdata=<post data>
    """

    if field and value:
        if postdata:
            raise SaltInvocationError(
                "Either a field and a value, or a chunk "
                "of POST data, may be specified, but not both."
            )
        postdata = {name.title(): {field: value}}

    if postdata is None:
        raise SaltInvocationError(
            "Either a field and a value, or a chunk of POST data must be specified."
        )

    status, result = _query(
        action=name,
        command=id_,
        method="POST",
        data=salt.utils.json.dumps(postdata),
    )
    return result


def show_artifact(id_):
    """
    Show an artifact

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.show_artifact <artifact id>
    """
    return show_item("artifact", id_)


def list_users():
    """
    List the users

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.list_users
    """
    return list_items("user")


def show_user(id_):
    """
    Show a user

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.show_user <user id>
    """
    return show_item("user", id_)


def update_user(id_, field, value):
    """
    Update a user

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.update_user <user id> <field> <new value>
    """
    return update_item("user", id_, field, value)


def query_user(query_string, order="UserName"):
    """
    Update a user

    CLI Example:

    .. code-block:: bash

        salt myminion rallydev.query_user '(Name contains Jo)'
    """
    return query_item("user", query_string, order)



import logging
import textwrap

import pytest

import salt.modules.beacons as beaconmod
import salt.modules.cp as cp
import salt.modules.pacmanpkg as pacmanpkg
import salt.modules.pkg_resource as pkg_resource
import salt.modules.yumpkg as yumpkg
import salt.states.beacon as beaconstate
import salt.states.pkg as pkg
import salt.utils.state as state_utils
from salt.loader.dunder import __opts__
from salt.utils.event import SaltEvent
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        cp: {
            "__opts__": __opts__.with_default(minion_opts),
        },
        pkg: {
            "__env__": "base",
            "__salt__": {},
            "__grains__": {"os": "CentOS", "os_family": "RedHat"},
            "__opts__": minion_opts,
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {"state.gen_tag": state_utils.gen_tag},
        },
        beaconstate: {
            "__salt__": {},
            "__opts__": minion_opts,
        },
        beaconmod: {
            "__salt__": {},
            "__opts__": minion_opts,
        },
        pacmanpkg: {
            "__salt__": {},
            "__opts__": minion_opts,
        },
        pkg_resource: {
            "__salt__": {},
            "__grains__": {"os": "CentOS", "os_family": "RedHat"},
        },
        yumpkg: {
            "__salt__": {},
            "__grains__": {"osarch": "x86_64", "osmajorrelease": 7},
            "__opts__": minion_opts,
        },
    }


@pytest.fixture(scope="module")
def pkgs():
    return {
        "pkga": {"old": "1.0.1", "new": "2.0.1"},
        "pkgb": {"old": "1.0.2", "new": "2.0.2"},
        "pkgc": {"old": "1.0.3", "new": "2.0.3"},
    }


@pytest.fixture(scope="module")
def list_pkgs():
    return {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
    }


def test_uptodate_with_changes(pkgs):
    """
    Test pkg.uptodate with simulated changes
    """

    list_upgrades = MagicMock(
        return_value={pkgname: pkgver["new"] for pkgname, pkgver in pkgs.items()}
    )
    upgrade = MagicMock(return_value=pkgs)
    version = MagicMock(side_effect=lambda pkgname, **_: pkgs[pkgname]["old"])

    with patch.dict(
        pkg.__salt__,
        {
            "pkg.list_upgrades": list_upgrades,
            "pkg.upgrade": upgrade,
            "pkg.version": version,
        },
    ):
        # Run state with test=false
        with patch.dict(pkg.__opts__, {"test": False}):
            ret = pkg.uptodate("dummy", test=True)
            assert ret["result"]
            assert ret["changes"] == pkgs

        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg.uptodate("dummy", test=True)
            assert ret["result"] is None
            assert ret["changes"] == pkgs


def test_uptodate_with_pkgs_with_changes(pkgs):
    """
    Test pkg.uptodate with simulated changes
    """

    list_upgrades = MagicMock(
        return_value={pkgname: pkgver["new"] for pkgname, pkgver in pkgs.items()}
    )
    upgrade = MagicMock(return_value=pkgs)
    version = MagicMock(side_effect=lambda pkgname, **_: pkgs[pkgname]["old"])

    with patch.dict(
        pkg.__salt__,
        {
            "pkg.list_upgrades": list_upgrades,
            "pkg.upgrade": upgrade,
            "pkg.version": version,
        },
    ):
        # Run state with test=false
        with patch.dict(pkg.__opts__, {"test": False}):
            ret = pkg.uptodate(
                "dummy",
                test=True,
                pkgs=[pkgname for pkgname in pkgs],
            )
            assert ret["result"]
            assert ret["changes"] == pkgs

        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg.uptodate(
                "dummy",
                test=True,
                pkgs=[pkgname for pkgname in pkgs],
            )
            assert ret["result"] is None
            assert ret["changes"] == pkgs


def test_uptodate_no_changes():
    """
    Test pkg.uptodate with no changes
    """
    list_upgrades = MagicMock(return_value={})
    upgrade = MagicMock(return_value={})

    with patch.dict(
        pkg.__salt__, {"pkg.list_upgrades": list_upgrades, "pkg.upgrade": upgrade}
    ):
        # Run state with test=false
        with patch.dict(pkg.__opts__, {"test": False}):
            ret = pkg.uptodate("dummy", test=True)
            assert ret["result"]
            assert ret["changes"] == {}

        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg.uptodate("dummy", test=True)
            assert ret["result"]
            assert ret["changes"] == {}


def test_uptodate_with_pkgs_no_changes(pkgs):
    """
    Test pkg.uptodate with no changes
    """

    list_upgrades = MagicMock(return_value={})
    upgrade = MagicMock(return_value={})

    with patch.dict(
        pkg.__salt__, {"pkg.list_upgrades": list_upgrades, "pkg.upgrade": upgrade}
    ):
        # Run state with test=false
        with patch.dict(pkg.__opts__, {"test": False}):
            ret = pkg.uptodate(
                "dummy",
                test=True,
                pkgs=[pkgname for pkgname in pkgs],
            )
            assert ret["result"]
            assert ret["changes"] == {}

        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg.uptodate(
                "dummy",
                test=True,
                pkgs=[pkgname for pkgname in pkgs],
            )
            assert ret["result"]
            assert ret["changes"] == {}


def test_uptodate_with_failed_changes(pkgs):
    """
    Test pkg.uptodate with simulated failed changes
    """

    list_upgrades = MagicMock(
        return_value={pkgname: pkgver["new"] for pkgname, pkgver in pkgs.items()}
    )
    upgrade = MagicMock(return_value={})
    version = MagicMock(side_effect=lambda pkgname, **_: pkgs[pkgname]["old"])

    with patch.dict(
        pkg.__salt__,
        {
            "pkg.list_upgrades": list_upgrades,
            "pkg.upgrade": upgrade,
            "pkg.version": version,
        },
    ):
        # Run state with test=false
        with patch.dict(pkg.__opts__, {"test": False}):
            ret = pkg.uptodate(
                "dummy",
                test=True,
                pkgs=[pkgname for pkgname in pkgs],
            )
            assert not ret["result"]
            assert ret["changes"] == {}

        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg.uptodate(
                "dummy",
                test=True,
                pkgs=[pkgname for pkgname in pkgs],
            )
            assert ret["result"] is None
            assert ret["changes"] == pkgs


@pytest.mark.parametrize(
    "version_string, expected_version_conditions",
    [
        (
            "> 1.0.0, < 15.0.0, != 14.0.1",
            [(">", "1.0.0"), ("<", "15.0.0"), ("!=", "14.0.1")],
        ),
        (
            "> 1.0.0,< 15.0.0,!= 14.0.1",
            [(">", "1.0.0"), ("<", "15.0.0"), ("!=", "14.0.1")],
        ),
        (">= 1.0.0, < 15.0.0", [(">=", "1.0.0"), ("<", "15.0.0")]),
        (">=1.0.0,< 15.0.0", [(">=", "1.0.0"), ("<", "15.0.0")]),
        ("< 15.0.0", [("<", "15.0.0")]),
        ("<15.0.0", [("<", "15.0.0")]),
        ("15.0.0", [("==", "15.0.0")]),
        ("", []),
    ],
)
def test_parse_version_string(version_string, expected_version_conditions):
    version_conditions = pkg._parse_version_string(version_string)
    assert len(expected_version_conditions) == len(version_conditions)
    for expected_version_condition, version_condition in zip(
        expected_version_conditions, version_conditions
    ):
        assert expected_version_condition[0] == version_condition[0]
        assert expected_version_condition[1] == version_condition[1]


@pytest.mark.parametrize(
    "version_string, installed_versions, expected_result",
    [
        ("> 1.0.0, < 15.0.0, != 14.0.1", [], False),
        ("> 1.0.0, < 15.0.0, != 14.0.1", ["1.0.0"], False),
        ("> 1.0.0, < 15.0.0, != 14.0.1", ["14.0.1"], False),
        ("> 1.0.0, < 15.0.0, != 14.0.1", ["16.0.0"], False),
        ("> 1.0.0, < 15.0.0, != 14.0.1", ["2.0.0"], True),
        (
            "> 1.0.0, < 15.0.0, != 14.0.1",
            ["1.0.0", "14.0.1", "16.0.0", "2.0.0"],
            True,
        ),
        ("> 15.0.0", [], False),
        ("> 15.0.0", ["1.0.0"], False),
        ("> 15.0.0", ["16.0.0"], True),
        ("15.0.0", [], False),
        ("15.0.0", ["15.0.0"], True),
        # No version specified, whatever version installed. This is threated like ANY version installed fulfills.
        ("", ["15.0.0"], True),
        # No version specified, no version installed.
        ("", [], False),
    ],
)
def test_fulfills_version_string(version_string, installed_versions, expected_result):
    msg = "version_string: {}, installed_versions: {}, expected_result: {}".format(
        version_string, installed_versions, expected_result
    )
    assert expected_result == pkg._fulfills_version_string(
        installed_versions, version_string
    )


@pytest.mark.parametrize(
    "installed_versions, operator, version, expected_result",
    [
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], "==", "1.0.0", True),
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], ">=", "1.0.0", True),
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], ">", "1.0.0", True),
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], "<", "2.0.0", True),
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], "<=", "2.0.0", True),
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], "!=", "1.0.0", True),
        (["1.0.0", "14.0.1", "16.0.0", "2.0.0"], "==", "17.0.0", False),
        (["1.0.0"], "!=", "1.0.0", False),
        ([], "==", "17.0.0", False),
    ],
)
def test_fulfills_version_spec(installed_versions, operator, version, expected_result):
    msg = (
        "installed_versions: {}, operator: {}, version: {}, expected_result: {}".format(
            installed_versions, operator, version, expected_result
        )
    )
    assert expected_result == pkg._fulfills_version_spec(
        installed_versions, operator, version
    )


def test_mod_beacon(tmp_path):
    """
    Test to create a beacon based on a pkg
    """
    name = "vim"

    with patch.dict(pkg.__salt__, {"beacons.list": MagicMock(return_value={})}):
        with patch.dict(pkg.__states__, {"beacon.present": beaconstate.present}):
            ret = pkg.mod_beacon(name, sfun="latest")
            expected = {
                "name": name,
                "changes": {},
                "result": False,
                "comment": (
                    "pkg.latest does not work with the mod_beacon state function"
                ),
            }

            assert ret == expected

            ret = pkg.mod_beacon(name, sfun="installed")
            expected = {
                "name": name,
                "changes": {},
                "result": True,
                "comment": "Not adding beacon.",
            }

            assert ret == expected

    event_returns = [
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_complete",
            "beacons": {},
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_complete",
            "beacons": {},
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacons_list_available_complete",
            "beacons": ["pkg"],
        },
        {
            "valid": True,
            "tag": "/salt/minion/minion_beacon_validation_complete",
            "vcomment": "Valid beacon configuration",
        },
        {
            "complete": True,
            "tag": "/salt/minion/minion_beacon_add_complete",
            "beacons": {
                "beacon_pkg_vim": [
                    {"pkgs": [name]},
                    {"interval": 60},
                    {"beacon_module": "pkg"},
                ]
            },
        },
    ]
    mock = MagicMock(return_value=True)
    beacon_state_mocks = {
        "beacons.list": beaconmod.list_,
        "beacons.add": beaconmod.add,
        "beacons.list_available": beaconmod.list_available,
        "event.fire": mock,
    }

    beacon_mod_mocks = {"event.fire": mock}

    sock_dir = str(tmp_path / "test-socks")
    with patch.dict(pkg.__states__, {"beacon.present": beaconstate.present}):
        with patch.dict(beaconstate.__salt__, beacon_state_mocks):
            with patch.dict(beaconmod.__salt__, beacon_mod_mocks):
                with patch.dict(
                    beaconmod.__opts__, {"beacons": {}, "sock_dir": sock_dir}
                ):
                    with patch.object(
                        SaltEvent, "get_event", side_effect=event_returns
                    ):
                        ret = pkg.mod_beacon(name, sfun="installed", beacon=True)
                        expected = {
                            "name": "beacon_pkg_vim",
                            "changes": {},
                            "result": True,
                            "comment": "Adding beacon_pkg_vim to beacons",
                        }

                        assert ret == expected


def test_mod_aggregate():
    """
    Test to mod_aggregate function
    """
    low = {
        "state": "pkg",
        "name": "other_pkgs",
        "pkgs": ["byobu"],
        "aggregate": True,
        "fun": "installed",
    }

    chunks = [
        {
            "state": "file",
            "name": "/tmp/install-vim",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "/tmp/install-vim",
            "order": 10000,
            "fun": "managed",
        },
        {
            "state": "file",
            "name": "/tmp/install-tmux",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "/tmp/install-tmux",
            "order": 10001,
            "fun": "managed",
        },
        {
            "state": "pkg",
            "name": "other_pkgs",
            "__sls__": "47628",
            "__env __": "base",
            "__id__": "other_pkgs",
            "pkgs": ["byobu"],
            "aggregate": True,
            "order": 10002,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "bc",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "bc",
            "hold": True,
            "order": 10003,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "vim",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "vim",
            "require": ["/tmp/install-vim"],
            "order": 10004,
            "fun": "installed",
        },
        {
            "state": "pkg",
            "name": "tmux",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "tmux",
            "require": ["/tmp/install-tmux"],
            "order": 10005,
            "fun": "installed",
        },
        {
            "state": "pkgrepo",
            "name": "deb https://packages.cloud.google.com/apt cloud-sdk main",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "google-cloud-repo",
            "humanname": "Google Cloud SDK",
            "file": "/etc/apt/sources.list.d/google-cloud-sdk.list",
            "key_url": "https://packages.cloud.google.com/apt/doc/apt-key.gpg",
            "order": 10006,
            "fun": "managed",
        },
        {
            "state": "pkg",
            "name": "google-cloud-sdk",
            "__sls__": "47628",
            "__env__": "base",
            "__id__": "google-cloud-sdk",
            "require": ["google-cloud-repo"],
            "order": 10007,
            "fun": "installed",
        },
    ]

    running = {
        "file_|-/tmp/install-vim_| -/tmp/install-vim_|-managed": {
            "changes": {},
            "comment": "File /tmp/install-vim exists with proper permissions. No changes made.",
            "name": "/tmp/install-vim",
            "result": True,
            "__sls__": "47628",
            "__run_num__": 0,
            "start_time": "18:41:20.987275",
            "duration": 5.833,
            "__id__": "/tmp/install-vim",
        },
        "file_|-/tmp/install-tmux_|-/tmp/install-tmux_|-managed": {
            "changes": {},
            "comment": "File /tmp/install-tmux exists with proper permissions. No changes made.",
            "name": "/tmp/install-tmux",
            "result": True,
            "__sls__": "47628",
            "__run_num__": 1,
            "start_time": "18:41:20.993258",
            "duration": 1.263,
            "__id__": "/tmp/install-tmux",
        },
    }

    expected = {
        "pkgs": ["byobu", "byobu", "vim", "tmux", "google-cloud-sdk"],
        "name": "other_pkgs",
        "fun": "installed",
        "aggregate": True,
        "state": "pkg",
    }
    res = pkg.mod_aggregate(low, chunks, running)
    assert res == expected


def test_installed_with_changes_test_true(list_pkgs):
    """
    Test pkg.installed with simulated changes
    """

    list_pkgs = MagicMock(return_value=list_pkgs)

    with patch.dict(
        pkg.__salt__,
        {
            "pkg.list_pkgs": list_pkgs,
        },
    ):
        expected = {"dummy": {"new": "installed", "old": ""}}
        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg.installed("dummy", test=True)
            assert ret["result"] is None
            assert ret["changes"] == expected


def test_installed_with_sources(list_pkgs, tmp_path):
    """
    Test pkg.installed with passing `sources`
    """

    list_pkgs = MagicMock(return_value=list_pkgs)
    pkg_source = tmp_path / "pkga-package-0.3.0.deb"

    with patch.dict(
        pkg.__salt__,
        {
            "cp.cache_file": cp.cache_file,
            "pkg.list_pkgs": list_pkgs,
            "pkg_resource.pack_sources": pkg_resource.pack_sources,
            "lowpkg.bin_pkg_info": MagicMock(),
        },
    ), patch("salt.fileclient.get_file_client", return_value=MagicMock()):
        try:
            ret = pkg.installed("install-pkgd", sources=[{"pkga": str(pkg_source)}])
            assert ret["result"] is False
        except TypeError as exc:
            if "got multiple values for keyword argument 'saltenv'" in str(exc):
                pytest.fail(f"TypeError should have not been raised: {exc}")
            raise exc from None


@pytest.mark.parametrize("action", ["removed", "purged"])
def test_removed_purged_with_changes_test_true(list_pkgs, action):
    """
    Test pkg.removed with simulated changes
    """

    list_pkgs = MagicMock(return_value=list_pkgs)

    mock_parse_targets = MagicMock(return_value=[{"pkga": None}, "repository"])

    with patch.dict(
        pkg.__salt__,
        {
            "pkg.list_pkgs": list_pkgs,
            "pkg_resource.parse_targets": mock_parse_targets,
            "pkg_resource.version_clean": MagicMock(return_value=None),
        },
    ):
        expected = {"pkga": {"new": f"{action}", "old": ""}}
        pkg_actions = {"removed": pkg.removed, "purged": pkg.purged}

        # Run state with test=true
        with patch.dict(pkg.__opts__, {"test": True}):
            ret = pkg_actions[action]("pkga", test=True)
            assert ret["result"] is None
            assert ret["changes"] == expected


@pytest.mark.parametrize(
    "package_manager",
    [("Zypper"), ("YUM/DNF"), ("APT")],
)
def test_held_unheld(package_manager):
    """
    Test pkg.held and pkg.unheld with Zypper, YUM/DNF and APT
    """

    if package_manager == "Zypper":
        list_holds_func = "pkg.list_locks"
        list_holds_mock = MagicMock(
            return_value={
                "bar": {
                    "type": "package",
                    "match_type": "glob",
                    "case_sensitive": "on",
                },
                "minimal_base": {
                    "type": "pattern",
                    "match_type": "glob",
                    "case_sensitive": "on",
                },
                "baz": {
                    "type": "package",
                    "match_type": "glob",
                    "case_sensitive": "on",
                },
            }
        )
    elif package_manager == "YUM/DNF":
        list_holds_func = "pkg.list_holds"
        list_holds_mock = MagicMock(
            return_value=[
                "bar-0:1.2.3-1.1.*",
                "baz-0:2.3.4-2.1.*",
            ]
        )
    elif package_manager == "APT":
        list_holds_func = "pkg.get_selections"
        list_holds_mock = MagicMock(
            return_value={
                "hold": [
                    "bar",
                    "baz",
                ]
            }
        )

    def pkg_hold(name, pkgs=None, *_args, **__kwargs):
        if name and pkgs is None:
            pkgs = [name]
        ret = {}
        for pkg in pkgs:
            ret.update(
                {
                    pkg: {
                        "name": pkg,
                        "changes": {"new": "hold", "old": ""},
                        "result": True,
                        "comment": f"Package {pkg} is now being held.",
                    }
                }
            )
        return ret

    def pkg_unhold(name, pkgs=None, *_args, **__kwargs):
        if name and pkgs is None:
            pkgs = [name]
        ret = {}
        for pkg in pkgs:
            ret.update(
                {
                    pkg: {
                        "name": pkg,
                        "changes": {"new": "", "old": "hold"},
                        "result": True,
                        "comment": f"Package {pkg} is no longer held.",
                    }
                }
            )
        return ret

    hold_mock = MagicMock(side_effect=pkg_hold)
    unhold_mock = MagicMock(side_effect=pkg_unhold)

    # Testing with Zypper
    with patch.dict(
        pkg.__salt__,
        {
            list_holds_func: list_holds_mock,
            "pkg.hold": hold_mock,
            "pkg.unhold": unhold_mock,
        },
    ):
        # Holding one of two packages
        ret = pkg.held("held-test", pkgs=["foo", "bar"])
        assert "foo" in ret["changes"]
        assert len(ret["changes"]) == 1
        hold_mock.assert_called_once_with(name="held-test", pkgs=["foo"])
        unhold_mock.assert_not_called()

        hold_mock.reset_mock()
        unhold_mock.reset_mock()

        # Holding one of two packages and replacing all the rest held packages
        ret = pkg.held("held-test", pkgs=["foo", "bar"], replace=True)
        assert "foo" in ret["changes"]
        assert "baz" in ret["changes"]
        assert len(ret["changes"]) == 2
        hold_mock.assert_called_once_with(name="held-test", pkgs=["foo"])
        unhold_mock.assert_called_once_with(name="held-test", pkgs=["baz"])

        hold_mock.reset_mock()
        unhold_mock.reset_mock()

        # Remove all holds
        ret = pkg.held("held-test", pkgs=[], replace=True)
        assert "bar" in ret["changes"]
        assert "baz" in ret["changes"]
        assert len(ret["changes"]) == 2
        hold_mock.assert_not_called()
        unhold_mock.assert_any_call(name="held-test", pkgs=["baz"])
        unhold_mock.assert_any_call(name="held-test", pkgs=["bar"])

        hold_mock.reset_mock()
        unhold_mock.reset_mock()

        # Unolding one of two packages
        ret = pkg.unheld("held-test", pkgs=["foo", "bar"])
        assert "bar" in ret["changes"]
        assert len(ret["changes"]) == 1
        unhold_mock.assert_called_once_with(name="held-test", pkgs=["bar"])
        hold_mock.assert_not_called()

        hold_mock.reset_mock()
        unhold_mock.reset_mock()

        # Remove all holds
        ret = pkg.unheld("held-test", all=True)
        assert "bar" in ret["changes"]
        assert "baz" in ret["changes"]
        assert len(ret["changes"]) == 2
        hold_mock.assert_not_called()
        unhold_mock.assert_any_call(name="held-test", pkgs=["baz"])
        unhold_mock.assert_any_call(name="held-test", pkgs=["bar"])


def test_installed_with_single_normalize():
    """
    Test pkg.installed with preventing multiple package name normalisation
    """

    list_no_weird_installed = {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
    }
    list_no_weird_installed_ver_list = {
        "pkga": ["1.0.1"],
        "pkgb": ["1.0.2"],
        "pkgc": ["1.0.3"],
    }
    list_with_weird_installed = {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
        "weird-name-1.2.3-1234.5.6.test7tst.x86_64": "20220214-2.1",
    }
    list_with_weird_installed_ver_list = {
        "pkga": ["1.0.1"],
        "pkgb": ["1.0.2"],
        "pkgc": ["1.0.3"],
        "weird-name-1.2.3-1234.5.6.test7tst.x86_64": ["20220214-2.1"],
    }
    list_pkgs = MagicMock(
        side_effect=[
            # For the package with version specified
            list_no_weird_installed_ver_list,
            {},
            list_no_weird_installed,
            list_no_weird_installed_ver_list,
            list_with_weird_installed,
            list_with_weird_installed_ver_list,
            # For the package with no version specified
            list_no_weird_installed_ver_list,
            {},
            list_no_weird_installed,
            list_no_weird_installed_ver_list,
            list_with_weird_installed,
            list_with_weird_installed_ver_list,
        ]
    )

    salt_dict = {
        "pkg.install": yumpkg.install,
        "pkg.list_pkgs": list_pkgs,
        "pkg.normalize_name": yumpkg.normalize_name,
        "pkg_resource.version_clean": pkg_resource.version_clean,
        "pkg_resource.parse_targets": pkg_resource.parse_targets,
    }

    with patch("salt.modules.yumpkg.list_pkgs", list_pkgs), patch(
        "salt.modules.yumpkg.version_cmp", MagicMock(return_value=0)
    ), patch(
        "salt.modules.yumpkg._call_yum", MagicMock(return_value={"retcode": 0})
    ) as call_yum_mock, patch.dict(
        pkg.__salt__, salt_dict
    ), patch.dict(
        pkg_resource.__salt__, salt_dict
    ), patch.dict(
        yumpkg.__salt__, salt_dict
    ), patch.dict(
        yumpkg.__grains__, {"os": "CentOS", "osarch": "x86_64", "osmajorrelease": 7}
    ), patch.object(
        yumpkg, "list_holds", MagicMock()
    ):
        expected = {
            "weird-name-1.2.3-1234.5.6.test7tst.x86_64": {
                "old": "",
                "new": "20220214-2.1",
            }
        }
        ret = pkg.installed(
            "test_install",
            pkgs=[{"weird-name-1.2.3-1234.5.6.test7tst.x86_64.noarch": "20220214-2.1"}],
        )
        call_yum_mock.assert_called_once()
        assert (
            "weird-name-1.2.3-1234.5.6.test7tst.x86_64-20220214-2.1"
            in call_yum_mock.mock_calls[0].args[0]
        )
        assert ret["result"]
        assert ret["changes"] == expected

        call_yum_mock.reset_mock()

        ret = pkg.installed(
            "test_install",
            pkgs=["weird-name-1.2.3-1234.5.6.test7tst.x86_64.noarch"],
        )
        call_yum_mock.assert_called_once()
        assert (
            "weird-name-1.2.3-1234.5.6.test7tst.x86_64"
            in call_yum_mock.mock_calls[0].args[0]
        )
        assert ret["result"]
        assert ret["changes"] == expected


def test_removed_with_single_normalize():
    """
    Test pkg.removed with preventing multiple package name normalisation
    """

    list_no_weird_installed = {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
    }
    list_no_weird_installed_ver_list = {
        "pkga": ["1.0.1"],
        "pkgb": ["1.0.2"],
        "pkgc": ["1.0.3"],
    }
    list_with_weird_installed = {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
        "weird-name-1.2.3-1234.5.6.test7tst.x86_64": "20220214-2.1",
    }
    list_with_weird_installed_ver_list = {
        "pkga": ["1.0.1"],
        "pkgb": ["1.0.2"],
        "pkgc": ["1.0.3"],
        "weird-name-1.2.3-1234.5.6.test7tst.x86_64": ["20220214-2.1"],
    }
    list_pkgs = MagicMock(
        side_effect=[
            # For the package with version specified
            list_with_weird_installed_ver_list,
            list_with_weird_installed,
            list_no_weird_installed,
            list_no_weird_installed_ver_list,
            # For the package with no version specified
            list_with_weird_installed_ver_list,
            list_with_weird_installed,
            list_no_weird_installed,
            list_no_weird_installed_ver_list,
        ]
    )

    salt_dict = {
        "pkg.remove": yumpkg.remove,
        "pkg.list_pkgs": list_pkgs,
        "pkg.normalize_name": yumpkg.normalize_name,
        "pkg_resource.parse_targets": pkg_resource.parse_targets,
        "pkg_resource.version_clean": pkg_resource.version_clean,
    }

    with patch("salt.modules.yumpkg.list_pkgs", list_pkgs), patch(
        "salt.modules.yumpkg.version_cmp", MagicMock(return_value=0)
    ), patch(
        "salt.modules.yumpkg._call_yum", MagicMock(return_value={"retcode": 0})
    ) as call_yum_mock, patch.dict(
        pkg.__salt__, salt_dict
    ), patch.dict(
        pkg_resource.__salt__, salt_dict
    ), patch.dict(
        yumpkg.__salt__, salt_dict
    ):
        expected = {
            "weird-name-1.2.3-1234.5.6.test7tst.x86_64": {
                "old": "20220214-2.1",
                "new": "",
            }
        }
        ret = pkg.removed(
            "test_remove",
            pkgs=[{"weird-name-1.2.3-1234.5.6.test7tst.x86_64.noarch": "20220214-2.1"}],
        )
        call_yum_mock.assert_called_once()
        assert (
            "weird-name-1.2.3-1234.5.6.test7tst.x86_64-20220214-2.1"
            in call_yum_mock.mock_calls[0].args[0]
        )
        assert ret["result"]
        assert ret["changes"] == expected

        call_yum_mock.reset_mock()

        ret = pkg.removed(
            "test_remove",
            pkgs=["weird-name-1.2.3-1234.5.6.test7tst.x86_64.noarch"],
        )
        call_yum_mock.assert_called_once()
        assert (
            "weird-name-1.2.3-1234.5.6.test7tst.x86_64"
            in call_yum_mock.mock_calls[0].args[0]
        )
        assert ret["result"]
        assert ret["changes"] == expected


def test_installed_with_single_normalize_32bit():
    """
    Test pkg.installed of 32bit package with preventing multiple package name normalisation
    """

    list_no_weird_installed = {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
    }
    list_no_weird_installed_ver_list = {
        "pkga": ["1.0.1"],
        "pkgb": ["1.0.2"],
        "pkgc": ["1.0.3"],
    }
    list_with_weird_installed = {
        "pkga": "1.0.1",
        "pkgb": "1.0.2",
        "pkgc": "1.0.3",
        "xz-devel.i686": "1.2.3",
    }
    list_with_weird_installed_ver_list = {
        "pkga": ["1.0.1"],
        "pkgb": ["1.0.2"],
        "pkgc": ["1.0.3"],
        "xz-devel.i686": ["1.2.3"],
    }
    list_pkgs = MagicMock(
        side_effect=[
            list_no_weird_installed_ver_list,
            {},
            list_no_weird_installed,
            list_no_weird_installed_ver_list,
            list_with_weird_installed,
            list_with_weird_installed,
            list_with_weird_installed_ver_list,
        ]
    )

    salt_dict = {
        "pkg.install": yumpkg.install,
        "pkg.list_pkgs": list_pkgs,
        "pkg.normalize_name": yumpkg.normalize_name,
        "pkg_resource.version_clean": pkg_resource.version_clean,
        "pkg_resource.parse_targets": pkg_resource.parse_targets,
    }

    with patch("salt.modules.yumpkg.list_pkgs", list_pkgs), patch(
        "salt.modules.yumpkg.version_cmp", MagicMock(return_value=0)
    ), patch(
        "salt.modules.yumpkg._call_yum", MagicMock(return_value={"retcode": 0})
    ) as call_yum_mock, patch.dict(
        pkg.__salt__, salt_dict
    ), patch.dict(
        pkg_resource.__salt__, salt_dict
    ), patch.dict(
        yumpkg.__salt__, salt_dict
    ), patch.dict(
        yumpkg.__grains__, {"os": "CentOS", "osarch": "x86_64", "osmajorrelease": 7}
    ):
        expected = {
            "xz-devel.i686": {
                "old": "",
                "new": "1.2.3",
            }
        }
        ret = pkg.installed(
            "test_install",
            pkgs=["xz-devel.i686"],
        )
        call_yum_mock.assert_called_once()
        assert "xz-devel.i686" in call_yum_mock.mock_calls[0].args[0]
        assert ret["result"]
        assert ret["changes"] == expected


@pytest.mark.parametrize(
    "kwargs, expected_cli_options",
    (
        (
            (
                "fromrepo=foo,bar",
                "someotherkwarg=test",
                "disablerepo=ignored",
                "enablerepo=otherignored",
                "disableexcludes=this_argument_is_also_ignored",
            ),
            ("--disablerepo=*", "--enablerepo=foo,bar"),
        ),
        (
            ("enablerepo=foo", "disablerepo=bar"),
            ("--disablerepo=bar", "--enablerepo=foo"),
        ),
        (
            ("disablerepo=foo",),
            ("--disablerepo=foo",),
        ),
        (
            ("enablerepo=bar",),
            ("--enablerepo=bar",),
        ),
    ),
)
def test_yumpkg_group_installed_with_repo_options(
    list_pkgs, kwargs, expected_cli_options
):
    """
    Test that running a pkg.group_installed with repo options on RPM-based
    systems results in the correct yum/dnf groupinfo command being run by
    pkg.group_info.
    """
    kwargs = dict(item.split("=", 1) for item in kwargs)
    run_stdout = MagicMock(
        return_value=textwrap.dedent(
            """\
        Group: MyGroup
         Group-Id: my-group
         Description: A test group
         Mandatory Packages:
            pkga
            pkgb
        """
        )
    )

    salt_dict = {
        "cmd.run_stdout": run_stdout,
        "pkg.group_diff": yumpkg.group_diff,
        "pkg.group_info": yumpkg.group_info,
    }

    name = "MyGroup"
    with patch.dict(pkg.__salt__, salt_dict), patch.dict(
        yumpkg.__salt__, salt_dict
    ), patch.object(
        yumpkg,
        "list_pkgs",
        MagicMock(return_value=list_pkgs),
    ):
        ret = pkg.group_installed(name, **kwargs)
        assert ret["result"]
        assert not ret["changes"]
        expected = [yumpkg._yum(), "--quiet"]
        expected.extend(expected_cli_options)
        expected.extend(("groupinfo", name))
        run_stdout.assert_called_once_with(
            expected,
            output_loglevel="trace",
            python_shell=False,
        )


def test_pacmanpkg_group_installed_with_repo_options(list_pkgs):
    """
    Test that running a pkg.group_installed with additional arguments on
    platforms which use pacman does not result in a traceback, but is instead
    cleanly handled and a useful comment included in the state return.
    """
    salt_dict = {
        "pkg.group_diff": pacmanpkg.group_diff,
    }

    with patch.dict(pkg.__salt__, salt_dict), patch.dict(pacmanpkg.__salt__, salt_dict):
        ret = pkg.group_installed("foo", fromrepo="bar")
        assert not ret["result"]
        assert not ret["changes"]
        assert ret["comment"] == "Repo options are not supported on this platform"


def test_latest():
    """
    Test pkg.latest
    """
    pkg_name = "fake_pkg"
    old_version = "1.2.2"
    new_version = "1.2.3"
    latest_version_mock = MagicMock(return_value={pkg_name: new_version})
    current_version_mock = MagicMock(return_value={pkg_name: old_version})
    install_mock = MagicMock(
        return_value={
            pkg_name: {
                "new": new_version,
                "old": old_version,
            },
        }
    )
    salt_dict = {
        "pkg.latest_version": latest_version_mock,
        "pkg.version": current_version_mock,
        "pkg.install": install_mock,
    }
    with patch.dict(pkg.__salt__, salt_dict):
        ret = pkg.latest(pkg_name)
        assert ret.get("result", False) is True


def test_latest_multiple_versions():
    """
    This case arises most often when updating the kernel, where multiple versions are now installed.

    See: https://github.com/saltstack/salt/issues/60931
    """
    pkg_name = "fake_pkg"
    old_version = "1.2.2"
    new_version = "1.2.3"
    latest_version_mock = MagicMock(return_value={pkg_name: new_version})
    current_version_mock = MagicMock(return_value={pkg_name: old_version})
    install_mock = MagicMock(
        return_value={
            pkg_name: {
                "new": f"{old_version},{new_version}",
                "old": old_version,
            },
        }
    )
    salt_dict = {
        "pkg.latest_version": latest_version_mock,
        "pkg.version": current_version_mock,
        "pkg.install": install_mock,
    }
    with patch.dict(pkg.__salt__, salt_dict):
        ret = pkg.latest(pkg_name)
        assert ret.get("result", False) is True


def test_latest_no_change_windows():
    """
    Test pkg.latest with no change to the package version for winrepo packages

    See: https://github.com/saltstack/salt/issues/65165
    """
    pkg_name = "fake_pkg"
    version = "1.2.2"
    latest_version_mock = MagicMock(return_value={pkg_name: version})
    current_version_mock = MagicMock(return_value={pkg_name: version})
    install_mock = MagicMock(return_value={pkg_name: {"install status": "success"}})
    salt_dict = {
        "pkg.latest_version": latest_version_mock,
        "pkg.version": current_version_mock,
        "pkg.install": install_mock,
    }
    with patch.dict(pkg.__salt__, salt_dict):
        ret = pkg.latest(pkg_name)
        assert ret.get("result", False) is True



import random
import string


def random_name(size=6):
    """
    Generates a random cloud instance name
    """
    return "CLOUD-TEST-" + "".join(
        random.choice(string.ascii_uppercase + string.digits) for x in range(size)
    )



"""
Installing of Windows features using DISM
=========================================

Install windows features/capabilties with DISM

.. code-block:: yaml

    Language.Basic~~~en-US~0.0.1.0:
      dism.capability_installed

    NetFx3:
      dism.feature_installed
"""

import logging
import os

import salt.utils.data
import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "dism"


def __virtual__():
    """
    Only work on Windows where the DISM module is available
    """
    if not salt.utils.platform.is_windows():
        return False, "Module only available on Windows"

    return __virtualname__


def capability_installed(
    name, source=None, limit_access=False, image=None, restart=False
):
    """
    Install a DISM capability

    Args:
        name (str): The capability to install
        source (str): The optional source of the capability
        limit_access (bool): Prevent DISM from contacting Windows Update for
            online images
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the install

    Example:
        Run ``dism.available_capabilities`` to get a list of available
        capabilities. This will help you get the proper name to use.

        .. code-block:: yaml

            install_dotnet35:
              dism.capability_installed:
                - name: NetFX3~~~~
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    old = __salt__["dism.installed_capabilities"]()

    if name in old:
        ret["comment"] = f"The capability {name} is already installed"
        return ret

    if __opts__["test"]:
        ret["changes"]["capability"] = f"{name} will be installed"
        ret["result"] = None
        return ret

    # Install the capability
    status = __salt__["dism.add_capability"](name, source, limit_access, image, restart)

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to install {}: {}".format(name, status["stdout"])
        ret["result"] = False

    new = __salt__["dism.installed_capabilities"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Installed {name}"
        ret["changes"] = status
        ret["changes"]["capability"] = changes

    return ret


def capability_removed(name, image=None, restart=False):
    """
    Uninstall a DISM capability

    Args:
        name (str): The capability to uninstall
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the
            uninstall

    Example:
        Run ``dism.installed_capabilities`` to get a list of installed
        capabilities. This will help you get the proper name to use.

        .. code-block:: yaml

            remove_dotnet35:
              dism.capability_removed:
                - name: NetFX3~~~~
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    old = __salt__["dism.installed_capabilities"]()

    if name not in old:
        ret["comment"] = f"The capability {name} is already removed"
        return ret

    if __opts__["test"]:
        ret["changes"]["capability"] = f"{name} will be removed"
        ret["result"] = None
        return ret

    # Remove the capability
    status = __salt__["dism.remove_capability"](name, image, restart)

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to remove {}: {}".format(name, status["stdout"])
        ret["result"] = False

    new = __salt__["dism.installed_capabilities"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Removed {name}"
        ret["changes"] = status
        ret["changes"]["capability"] = changes

    return ret


def feature_installed(
    name,
    package=None,
    source=None,
    limit_access=False,
    enable_parent=False,
    image=None,
    restart=False,
):
    """
    Install a DISM feature

    Args:
        name (str): The feature in which to install
        package (Optional[str]): The parent package for the feature. You do not
            have to specify the package if it is the Windows Foundation Package.
            Otherwise, use package to specify the parent package of the feature
        source (str): The optional source of the feature
        limit_access (bool): Prevent DISM from contacting Windows Update for
            online images
        enable_parent (Optional[bool]): True will enable all parent features of
            the specified feature
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the install

    Example:
        Run ``dism.available_features`` to get a list of available features.
        This will help you get the proper name to use.

        .. code-block:: yaml

            install_telnet_client:
              dism.feature_installed:
                - name: TelnetClient
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    old = __salt__["dism.installed_features"]()

    if name in old:
        ret["comment"] = f"The feature {name} is already installed"
        return ret

    if __opts__["test"]:
        ret["changes"]["feature"] = f"{name} will be installed"
        ret["result"] = None
        return ret

    # Install the feature
    status = __salt__["dism.add_feature"](
        name, package, source, limit_access, enable_parent, image, restart
    )

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to install {}: {}".format(name, status["stdout"])
        ret["result"] = False

    new = __salt__["dism.installed_features"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Installed {name}"
        ret["changes"] = status
        ret["changes"]["feature"] = changes

    return ret


def feature_removed(name, remove_payload=False, image=None, restart=False):
    """
    Disables a feature.

    Args:
        name (str): The feature to disable
        remove_payload (Optional[bool]): Remove the feature's payload. Must
            supply source when enabling in the future.
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the
            uninstall

    Example:
        Run ``dism.installed_features`` to get a list of installed features.
        This will help you get the proper name to use.

        .. code-block:: yaml

            remove_telnet_client:
              dism.feature_removed:
                - name: TelnetClient
                - remove_payload: True
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    old = __salt__["dism.installed_features"]()

    if name not in old:
        ret["comment"] = f"The feature {name} is already removed"
        return ret

    if __opts__["test"]:
        ret["changes"]["feature"] = f"{name} will be removed"
        ret["result"] = None
        return ret

    # Remove the feature
    status = __salt__["dism.remove_feature"](name, remove_payload, image, restart)

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to remove {}: {}".format(name, status["stdout"])
        ret["result"] = False

    new = __salt__["dism.installed_features"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Removed {name}"
        ret["changes"] = status
        ret["changes"]["feature"] = changes

    return ret


def package_installed(
    name, ignore_check=False, prevent_pending=False, image=None, restart=False
):
    """
    Install a package.

    Args:
        name (str): The package to install. Can be a .cab file, a .msu file,
            or a folder
        ignore_check (Optional[bool]): Skip installation of the package if the
            applicability checks fail
        prevent_pending (Optional[bool]): Skip the installation of the package
            if there are pending online actions
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the install

    Example:

        .. code-block:: yaml

            install_KB123123123:
              dism.package_installed:
                - name: C:\\Packages\\KB123123123.cab
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    # Fail if using a non-existent package path
    if "~" not in name and not os.path.exists(name):
        if __opts__["test"]:
            ret["result"] = None
        else:
            ret["result"] = False
        ret["comment"] = f"Package path {name} does not exist"
        return ret

    old = __salt__["dism.installed_packages"]()

    # Get package info so we can see if it's already installed
    package_info = __salt__["dism.package_info"](name)

    if package_info["Package Identity"] in old:
        ret["comment"] = "The package {} is already installed: {}".format(
            name, package_info["Package Identity"]
        )
        return ret

    if __opts__["test"]:
        ret["changes"]["package"] = f"{name} will be installed"
        ret["result"] = None
        return ret

    # Install the package
    status = __salt__["dism.add_package"](
        name, ignore_check, prevent_pending, image, restart
    )

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to install {}: {}".format(name, status["stdout"])
        ret["result"] = False

    new = __salt__["dism.installed_packages"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Installed {name}"
        ret["changes"] = status
        ret["changes"]["package"] = changes

    return ret


def package_removed(name, image=None, restart=False):
    """
    Uninstall a package

    Args:
        name (str): The full path to the package. Can be either a .cab file or a
            folder. Should point to the original source of the package, not to
            where the file is installed. This can also be the name of a package as listed in
            ``dism.installed_packages``
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the
            uninstall

    Example:

        .. code-block:: yaml

            # Example using source
            remove_KB1231231:
              dism.package_installed:
                - name: C:\\Packages\\KB1231231.cab

            # Example using name from ``dism.installed_packages``
            remove_KB1231231:
              dism.package_installed:
                - name: Package_for_KB1231231~31bf3856ad364e35~amd64~~10.0.1.3
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    # Fail if using a non-existent package path
    if "~" not in name and not os.path.exists(name):
        if __opts__["test"]:
            ret["result"] = None
        else:
            ret["result"] = False
        ret["comment"] = f"Package path {name} does not exist"
        return ret

    old = __salt__["dism.installed_packages"]()

    # Get package info so we can see if it's already removed
    package_info = __salt__["dism.package_info"](name)

    # If `Package Identity` isn't returned or if they passed a cab file, if
    # `Package Identity` isn't in the list of installed packages
    if (
        "Package Identity" not in package_info
        or package_info["Package Identity"] not in old
    ):
        ret["comment"] = f"The package {name} is already removed"
        return ret

    if __opts__["test"]:
        ret["changes"]["package"] = f"{name} will be removed"
        ret["result"] = None
        return ret

    # Remove the package
    status = __salt__["dism.remove_package"](name, image, restart)

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to remove {}: {}".format(name, status["stdout"])
        ret["result"] = False

    new = __salt__["dism.installed_packages"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Removed {name}"
        ret["changes"] = status
        ret["changes"]["package"] = changes

    return ret


def kb_removed(name, image=None, restart=False):
    """
    Uninstall a KB package

    .. versionadded:: 3006.0

    Args:
        name (str): The name of the KB. Can be with or without the KB at the
            beginning.
        image (Optional[str]): The path to the root directory of an offline
            Windows image. If `None` is passed, the running operating system is
            targeted. Default is None.
        restart (Optional[bool]): Reboot the machine if required by the
            uninstall

    Example:

        .. code-block:: yaml

            # Example using full KB name
            remove_KB1231231:
              dism.package_installed:
                - name: KB1231231

            # Example using just he KB number
            remove_KB1231231:
              dism.package_installed:
                - name: 1231231
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    pkg_name = __salt__["dism.get_kb_package_name"](kb=name, image=image)

    # If pkg_name is None, the package is not installed
    if pkg_name is None:
        ret["comment"] = f"{name} is not installed"
        return ret

    if __opts__["test"]:
        ret["changes"]["package"] = f"{name} will be removed"
        ret["result"] = None
        return ret

    # Fail if using a non-existent package path
    old = __salt__["dism.installed_packages"]()

    # Remove the package
    status = __salt__["dism.remove_kb"](kb=name, image=image, restart=restart)

    if status["retcode"] not in [0, 1641, 3010]:
        ret["comment"] = "Failed to remove {}: {}".format(name, status["stdout"])
        ret["result"] = False
        return ret

    new = __salt__["dism.installed_packages"]()
    changes = salt.utils.data.compare_lists(old, new)

    if changes:
        ret["comment"] = f"Removed {name}"
        ret["changes"] = status
        ret["changes"]["package"] = changes

    return ret



import pytest

from salt.utils import win_runas


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ("test_user", ("test_user", ".")),  # Simple system name
        ("domain\\test_user", ("test_user", "domain")),  # Sam name
        ("domain.com\\test_user", ("test_user", "domain.com")),  # Sam name with .com
        ("test_user@domain", ("test_user", "domain")),  # UPN Name
        ("test_user@domain.com", ("test_user", "domain.com")),  # UPN Name with .com
        ("test_user@domain.local", ("test_user", "domain")),  # UPN Name with .local
    ],
)
def test_split_username(input_value, expected):
    """
    Test that the username is parsed properly from various domain/username
    combinations
    """
    result = win_runas.split_username(input_value)
    assert result == expected



import stat
from textwrap import dedent

import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account() as _account:
        yield _account


@pytest.fixture
def echo_script(state_tree):
    if salt.utils.platform.is_windows():
        file_name = "echo_script.bat"
        contents = dedent(
            """\
            @echo off
            set a=%~1
            set b=%~2
            echo a: %a%, b: %b%
            """
        )
    else:
        file_name = "echo_script.sh"
        contents = dedent(
            """\
            #!/bin/bash
            a="$1"
            b="$2"
            echo "a: $a, b: $b"
            """
        )
    with pytest.helpers.temp_file(file_name, contents, state_tree / "echo-script"):
        yield file_name


@pytest.fixture
def pipe_script(state_tree):
    if salt.utils.platform.is_windows():
        file_name = "pipe_script.bat"
        contents = dedent(
            """\
            @echo off
            IF "%1" == "|" (
                echo b0rken
            ) ELSE (
                echo fine
            )
            """
        )
    else:
        file_name = "pipe_script.sh"
        contents = dedent(
            """\
            #!/bin/bash
            if [ "$1" == '|' ]; then
                echo b0rken
            else
                echo fine
            fi
            """
        )
    with pytest.helpers.temp_file(file_name, contents, state_tree / "echo-script") as f:
        if not salt.utils.platform.is_windows():
            current_perms = f.stat().st_mode
            new_perms = current_perms | stat.S_IXUSR
            f.chmod(new_perms)
            f.chmod(0o755)
        yield f


@pytest.mark.parametrize(
    "args, expected",
    [
        ("foo bar", "a: foo, b: bar"),
        ('foo "bar bar"', "a: foo, b: bar bar"),
        (["foo", "bar"], "a: foo, b: bar"),
        (["foo foo", "bar bar"], "a: foo foo, b: bar bar"),
    ],
)
def test_echo(modules, echo_script, args, expected):
    """
    Test argument processing with a batch script
    """
    script = f"salt://echo-script/{echo_script}"
    result = modules.cmd.script(script, args=args)
    assert result["stdout"] == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        ("foo bar", "a: foo, b: bar"),
        ('foo "bar bar"', "a: foo, b: bar bar"),
        (["foo", "bar"], "a: foo, b: bar"),
        (["foo foo", "bar bar"], "a: foo foo, b: bar bar"),
    ],
)
def test_echo_runas(modules, account, echo_script, args, expected):
    """
    Test argument processing with a batch/bash script and runas
    """
    script = f"salt://echo-script/{echo_script}"
    result = modules.cmd.script(
        script,
        args=args,
        runas=account.username,
        password=account.password,
    )
    assert result["stdout"] == expected


def test_pipe_run_python_shell_true(modules, pipe_script):
    if salt.utils.platform.is_windows():
        cmd = f'{str(pipe_script)} | find /c /v ""'
    else:
        cmd = f"{str(pipe_script)} | wc -l"
    result = modules.cmd.run(cmd, python_shell=True)
    assert result == "1"


def test_pipe_run_python_shell_false(modules, pipe_script):
    if salt.utils.platform.is_windows():
        cmd = f'{str(pipe_script)} | find /c /v ""'
        # Behavior is different on Windows, I think it has to do with how cmd
        # deals with args vs bash... or maybe how args are passed on Windows
        expected = "1"
    else:
        cmd = f"{str(pipe_script)} | wc -l"
        expected = "b0rken"
    result = modules.cmd.run(cmd, python_shell=False)
    assert result == expected


def test_pipe_run_default(modules, pipe_script):
    if salt.utils.platform.is_windows():
        cmd = f'{str(pipe_script)} | find /c /v ""'
    else:
        cmd = f"{str(pipe_script)} | wc -l"
    # We need to mock running from the CLI by passing __pub_jid
    # Normally this is populated when run from the CLI, but when run from the
    # test suite, the value is empty
    result = modules.cmd.run(cmd, __pub_jid="test")
    assert result == "1"


def test_pipe_run_shell(modules, pipe_script):
    if salt.utils.platform.is_windows():
        cmd = f'{str(pipe_script)} | find /c /v ""'
        shell = "cmd"
    else:
        cmd = f"{str(pipe_script)} | wc -l"
        shell = "/bin/bash"
    # We need to mock running from the CLI by passing __pub_jid
    # Normally this is populated when run from the CLI, but when run from the
    # test suite, the value is empty
    result = modules.cmd.run(cmd, shell=shell, __pub_jid="test")
    assert result == "1"



"""
Tests for the Azure Blob External Pillar.
"""

import pickle
import time

import pytest

import salt.config
import salt.loader
import salt.pillar.azureblob as azureblob
import salt.utils.files
from tests.support.mock import MagicMock, patch

HAS_LIBS = False
try:
    # pylint: disable=no-name-in-module
    from azure.storage.blob import BlobServiceClient

    # pylint: enable=no-name-in-module

    HAS_LIBS = True
except ImportError:
    pass


pytestmark = [
    pytest.mark.skipif(
        HAS_LIBS is False,
        reason="The azure.storage.blob module must be installed.",
    )
]


class MockBlob(dict):
    """
    Creates a Mock Blob object.
    """

    name = ""

    def __init__(self):
        super().__init__(
            {
                "container": None,
                "name": "test.sls",
                "prefix": None,
                "delimiter": "/",
                "results_per_page": None,
                "location_mode": None,
            }
        )


class MockContainerClient:
    """
    Creates a Mock ContainerClient.
    """

    def __init__(self):
        pass

    def walk_blobs(self, *args, **kwargs):
        yield MockBlob()

    def get_blob_client(self, *args, **kwargs):
        pass


class MockBlobServiceClient:
    """
    Creates a Mock BlobServiceClient.
    """

    def __init__(self):
        pass

    def get_container_client(self, *args, **kwargs):
        container_client = MockContainerClient()
        return container_client


@pytest.fixture
def cachedir(tmp_path):
    dirname = tmp_path / "cachedir"
    dirname.mkdir(parents=True, exist_ok=True)
    return dirname


@pytest.fixture
def configure_loader_modules(cachedir, tmp_path):
    base_pillar = tmp_path / "base"
    prod_pillar = tmp_path / "prod"
    base_pillar.mkdir(parents=True, exist_ok=True)
    prod_pillar.mkdir(parents=True, exist_ok=True)
    pillar_roots = {
        "base": [str(base_pillar)],
        "prod": [str(prod_pillar)],
    }
    opts = {
        "cachedir": cachedir,
        "pillar_roots": pillar_roots,
    }
    return {
        azureblob: {"__opts__": opts},
    }


def test__init_expired(tmp_path):
    """
    Tests the result of _init when the cache is expired.
    """
    container = "test"
    multiple_env = False
    environment = "base"
    blob_cache_expire = 0  # The cache will be expired
    blob_client = MockBlobServiceClient()
    cache_file = tmp_path / "cache_file"
    # Patches the _get_containers_cache_filename module so that it returns the name of the new tempfile that
    # represents the cache file
    with patch.object(
        azureblob,
        "_get_containers_cache_filename",
        MagicMock(return_value=str(cache_file)),
    ):
        # Patches the from_connection_string module of the BlobServiceClient class so that a connection string does
        # not need to be given. Additionally it returns example blob data used by the ext_pillar.
        with patch.object(
            BlobServiceClient,
            "from_connection_string",
            MagicMock(return_value=blob_client),
        ):
            ret = azureblob._init(
                "", container, multiple_env, environment, blob_cache_expire
            )

        expected = {
            "base": {
                "test": [
                    {
                        "container": None,
                        "name": "test.sls",
                        "prefix": None,
                        "delimiter": "/",
                        "results_per_page": None,
                        "location_mode": None,
                    }
                ]
            }
        }
        assert ret == expected


def test__init_not_expired(tmp_path):
    """
    Tests the result of _init when the cache is not expired.
    """
    container = "test"
    multiple_env = False
    environment = "base"
    blob_cache_expire = (time.time()) * (time.time())  # The cache will not be expired
    metadata = {
        "base": {
            "test": [
                {"name": "base/secret.sls", "relevant": "include.sls"},
                {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
            ]
        }
    }
    cache_file = tmp_path / "cache_file"
    # Pickles the metadata and stores it in cache_file
    with salt.utils.files.fopen(str(cache_file), "wb") as fp_:
        pickle.dump(metadata, fp_)
    # Patches the _get_containers_cache_filename module so that it returns the name of the new tempfile that
    # represents the cache file
    with patch.object(
        azureblob,
        "_get_containers_cache_filename",
        MagicMock(return_value=str(cache_file)),
    ):
        # Patches the _read_containers_cache_file module so that it returns what it normally would if the new
        # tempfile representing the cache file was passed to it
        plugged = azureblob._read_containers_cache_file(str(cache_file))
        with patch.object(
            azureblob,
            "_read_containers_cache_file",
            MagicMock(return_value=plugged),
        ):
            ret = azureblob._init(
                "", container, multiple_env, environment, blob_cache_expire
            )
    assert ret == metadata


def test__get_cache_dir(cachedir):
    """
    Tests the result of _get_cache_dir.
    """
    ret = azureblob._get_cache_dir()
    assert ret == str(cachedir / "pillar_azureblob")


def test__get_cached_file_name(cachedir):
    """
    Tests the result of _get_cached_file_name.
    """
    container = "test"
    saltenv = "base"
    path = "base/secret.sls"
    ret = azureblob._get_cached_file_name(container, saltenv, path)
    assert ret == str(cachedir / "pillar_azureblob" / saltenv / container / path)


def test__get_containers_cache_filename(cachedir):
    """
    Tests the result of _get_containers_cache_filename.
    """
    container = "test"
    ret = azureblob._get_containers_cache_filename(container)
    assert ret == str(cachedir / "pillar_azureblob" / "test-files.cache")


def test__refresh_containers_cache_file(tmp_path):
    """
    Tests the result of _refresh_containers_cache_file to ensure that it successfully copies blob data into a
        cache file.
    """
    blob_client = MockBlobServiceClient()
    container = "test"
    cache_file = tmp_path / "cache_file"
    with patch.object(
        BlobServiceClient,
        "from_connection_string",
        MagicMock(return_value=blob_client),
    ):
        ret = azureblob._refresh_containers_cache_file("", container, str(cache_file))
    expected = {
        "base": {
            "test": [
                {
                    "container": None,
                    "name": "test.sls",
                    "prefix": None,
                    "delimiter": "/",
                    "results_per_page": None,
                    "location_mode": None,
                }
            ]
        }
    }
    assert ret == expected


def test__read_containers_cache_file(tmp_path):
    """
    Tests the result of _read_containers_cache_file to make sure that it successfully loads in pickled metadata.
    """
    metadata = {
        "base": {
            "test": [
                {"name": "base/secret.sls", "relevant": "include.sls"},
                {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
            ]
        }
    }
    cache_file = tmp_path / "cache_file"
    # Pickles the metadata and stores it in cache_file
    with salt.utils.files.fopen(str(cache_file), "wb") as fp_:
        pickle.dump(metadata, fp_)
    # Checks to see if _read_containers_cache_file can successfully read the pickled metadata from the cache file
    ret = azureblob._read_containers_cache_file(str(cache_file))
    assert ret == metadata


def test__find_files():
    """
    Tests the result of _find_files. Ensures it only finds files and not directories. Ensures it also ignore
        irrelevant files.
    """
    metadata = {
        "test": [
            {"name": "base/secret.sls"},
            {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
            {"name": "base/"},
        ]
    }
    ret = azureblob._find_files(metadata)
    assert ret == {"test": ["base/secret.sls", "blobtest.sls"]}


def test__find_file_meta1():
    """
    Tests the result of _find_file_meta when the metadata contains a blob with the specified path and a blob
        without the specified path.
    """
    metadata = {
        "base": {
            "test": [
                {"name": "base/secret.sls", "relevant": "include.sls"},
                {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
            ]
        }
    }
    container = "test"
    saltenv = "base"
    path = "base/secret.sls"
    ret = azureblob._find_file_meta(metadata, container, saltenv, path)
    assert ret == {"name": "base/secret.sls", "relevant": "include.sls"}


def test__find_file_meta2():
    """
    Tests the result of _find_file_meta when the saltenv in metadata does not match the specified saltenv.
    """
    metadata = {"wrong": {"test": [{"name": "base/secret.sls"}]}}
    container = "test"
    saltenv = "base"
    path = "base/secret.sls"
    ret = azureblob._find_file_meta(metadata, container, saltenv, path)
    assert ret is None


def test__find_file_meta3():
    """
    Tests the result of _find_file_meta when the container in metadata does not match the specified metadata.
    """
    metadata = {"base": {"wrong": [{"name": "base/secret.sls"}]}}
    container = "test"
    saltenv = "base"
    path = "base/secret.sls"
    ret = azureblob._find_file_meta(metadata, container, saltenv, path)
    assert ret is None



"""
Ensure pillar overrides are merged recursively, that wrapper
modules are in sync with the pillar dict in the rendering environment
and that the pillars are available on the target.
"""

import json

import pytest

import salt.utils.dictupdate

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.usefixtures("pillar_tree_nested"),
    pytest.mark.slow_test,
]


def test_pillar_is_only_rendered_once_without_overrides(salt_ssh_cli, caplog):
    ret = salt_ssh_cli.run("state.apply", "test")
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data[next(iter(ret.data))]["result"] is True
    assert caplog.text.count("hithere: pillar was rendered") == 1


def test_pillar_is_rerendered_with_overrides(salt_ssh_cli, caplog):
    ret = salt_ssh_cli.run("state.apply", "test", pillar={"foo": "bar"})
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    assert ret.data[next(iter(ret.data))]["result"] is True
    assert caplog.text.count("hithere: pillar was rendered") == 2


@pytest.fixture(scope="module", autouse=True)
def _show_pillar_state(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - showpillar
      '127.0.0.1':
        - showpillar
    """
    show_pillar_sls = """
    deep_thought:
      test.show_notification:
        - text: '{{ {
            "raw": {
              "the_meaning": pillar.get("the_meaning"),
              "btw": pillar.get("btw")},
            "wrapped": {
              "the_meaning": salt["pillar.get"]("the_meaning"),
              "btw": salt["pillar.get"]("btw")}}
            | json }}'

    target_check:
      test.check_pillar:
        - present:
          - the_meaning:of:foo
          - btw
          - the_meaning:of:bar
          - the_meaning:for
        - listing:
          - the_meaning:of:life
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    show_tempfile = pytest.helpers.temp_file(
        "showpillar.sls", show_pillar_sls, base_env_state_tree_root_dir
    )
    with top_tempfile, show_tempfile:
        yield


@pytest.fixture
def base():
    return {"the_meaning": {"of": {"life": 42, "bar": "tender"}, "for": "what"}}


@pytest.fixture
def override(base):
    poverride = {
        "the_meaning": {"of": {"life": [2.71], "foo": "lish"}},
        "btw": "turtles",
    }
    expected = salt.utils.dictupdate.merge(base, poverride)
    return expected, poverride


def test_state_sls(salt_ssh_cli, override):
    expected, override = override
    ret = salt_ssh_cli.run("state.sls", "showpillar", pillar=override)
    _assert_basic(ret)
    assert len(ret.data) == 2
    for sid, sret in ret.data.items():
        if "show" in sid:
            _assert_pillar(sret["comment"], expected)
        else:
            assert sret["result"] is True


@pytest.mark.parametrize("sid", ("deep_thought", "target_check"))
def test_state_sls_id(salt_ssh_cli, sid, override):
    expected, override = override
    ret = salt_ssh_cli.run("state.sls_id", sid, "showpillar", pillar=override)
    _assert_basic(ret)
    state_res = ret.data[next(iter(ret.data))]
    if sid == "deep_thought":
        _assert_pillar(state_res["comment"], expected)
    else:
        assert state_res["result"] is True


def test_state_highstate(salt_ssh_cli, override):
    expected, override = override
    ret = salt_ssh_cli.run("state.highstate", pillar=override, whitelist=["showpillar"])
    _assert_basic(ret)
    assert len(ret.data) == 2
    for sid, sret in ret.data.items():
        if "show" in sid:
            _assert_pillar(sret["comment"], expected)
        else:
            assert sret["result"] is True


def test_state_show_sls(salt_ssh_cli, override):
    expected, override = override
    ret = salt_ssh_cli.run("state.show_sls", "showpillar", pillar=override)
    _assert_basic(ret)
    pillar = ret.data["deep_thought"]["test"]
    pillar = next(x["text"] for x in pillar if isinstance(x, dict))
    _assert_pillar(pillar, expected)


def test_state_show_low_sls(salt_ssh_cli, override):
    expected, override = override
    ret = salt_ssh_cli.run("state.show_low_sls", "showpillar", pillar=override)
    _assert_basic(ret, list)
    pillar = ret.data[0]["text"]
    _assert_pillar(pillar, expected)


def test_state_single(salt_ssh_cli, override):
    expected, override = override
    ret = salt_ssh_cli.run(
        "state.single",
        "test.check_pillar",
        "foo",
        present=[
            "the_meaning:of:foo",
            "btw",
            "the_meaning:of:bar",
            "the_meaning:for",
        ],
        listing=["the_meaning:of:life"],
        pillar=override,
    )
    _assert_basic(ret, dict)
    state_res = ret.data[next(iter(ret.data))]
    assert state_res["result"] is True


def test_state_top(salt_ssh_cli, override):
    expected, override = override
    ret = salt_ssh_cli.run("state.top", "top.sls", pillar=override)
    _assert_basic(ret)
    assert len(ret.data) == 2
    for sid, sret in ret.data.items():
        if "show" in sid:
            _assert_pillar(sret["comment"], expected)
        else:
            assert sret["result"] is True


def _assert_pillar(pillar, expected):
    if not isinstance(pillar, dict):
        pillar = json.loads(pillar)
    assert pillar["raw"] == expected
    assert pillar["wrapped"] == expected


def _assert_basic(ret, typ=dict):
    assert ret.returncode == 0
    assert isinstance(ret.data, typ)
    assert ret.data



import hashlib
import logging
import os
import shutil
import signal
import tempfile
import textwrap
import time
import uuid

import psutil  # pylint: disable=3rd-party-module-not-gated
import pytest
from pytestshellutils.utils import ports
from saltfactories.utils.tempfiles import temp_file

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from tests.support.case import ModuleCase
from tests.support.helpers import with_tempfile
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.mark.windows_whitelisted
class CPModuleTest(ModuleCase):
    """
    Validate the cp module
    """

    def run_function(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """
        Ensure that results are decoded

        TODO: maybe move this behavior to ModuleCase itself?
        """
        return salt.utils.data.decode(super().run_function(*args, **kwargs))

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_file(self, tgt):
        """
        cp.get_file
        """
        self.run_function("cp.get_file", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_get_file_to_dir(self):
        """
        cp.get_file
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "")
        self.run_function("cp.get_file", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt + "scene33", "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @pytest.mark.skip_on_windows(reason="This test hangs on Windows on Py3")
    def test_get_file_templated_paths(self, tgt):
        """
        cp.get_file
        """
        self.run_function(
            "cp.get_file",
            [
                "salt://{{grains.test_grain}}",
                tgt.replace("cheese", "{{grains.test_grain}}"),
            ],
            template="jinja",
        )
        with salt.utils.files.fopen(tgt, "r") as cheese:
            data = salt.utils.stringutils.to_unicode(cheese.read())
        self.assertIn("Gromit", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_file_gzipped(self, tgt):
        """
        cp.get_file
        """
        src = os.path.join(RUNTIME_VARS.FILES, "file", "base", "file.big")
        with salt.utils.files.fopen(src, "rb") as fp_:
            hash_str = hashlib.sha256(fp_.read()).hexdigest()

        self.run_function("cp.get_file", ["salt://file.big", tgt], gzip=5)
        with salt.utils.files.fopen(tgt, "rb") as scene:
            data = scene.read()
        self.assertEqual(hash_str, hashlib.sha256(data).hexdigest())
        data = salt.utils.stringutils.to_unicode(data)
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_get_file_makedirs(self):
        """
        cp.get_file
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "make", "dirs", "scene33")
        self.run_function("cp.get_file", ["salt://grail/scene33", tgt], makedirs=True)
        self.addCleanup(
            shutil.rmtree, os.path.join(RUNTIME_VARS.TMP, "make"), ignore_errors=True
        )
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_template(self, tgt):
        """
        cp.get_template
        """
        self.run_function(
            "cp.get_template", ["salt://grail/scene33", tgt], spam="bacon"
        )
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("bacon", data)
        self.assertNotIn("spam", data)

    @pytest.mark.slow_test
    def test_get_dir(self):
        """
        cp.get_dir
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "many")
        self.run_function("cp.get_dir", ["salt://grail", tgt])
        self.assertIn("grail", os.listdir(tgt))
        self.assertIn("36", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("empty", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("scene", os.listdir(os.path.join(tgt, "grail", "36")))

    @pytest.mark.slow_test
    def test_get_dir_templated_paths(self):
        """
        cp.get_dir
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "many")
        self.run_function(
            "cp.get_dir",
            ["salt://{{grains.script}}", tgt.replace("many", "{{grains.alot}}")],
        )
        self.assertIn("grail", os.listdir(tgt))
        self.assertIn("36", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("empty", os.listdir(os.path.join(tgt, "grail")))
        self.assertIn("scene", os.listdir(os.path.join(tgt, "grail", "36")))

    # cp.get_url tests

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_url(self, tgt):
        """
        cp.get_url with salt:// source given
        """
        self.run_function("cp.get_url", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_get_url_makedirs(self):
        """
        cp.get_url
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "make", "dirs", "scene33")
        self.run_function("cp.get_url", ["salt://grail/scene33", tgt], makedirs=True)
        self.addCleanup(
            shutil.rmtree, os.path.join(RUNTIME_VARS.TMP, "make"), ignore_errors=True
        )
        with salt.utils.files.fopen(tgt, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_get_url_dest_empty(self):
        """
        cp.get_url with salt:// source given and destination omitted.
        """
        ret = self.run_function("cp.get_url", ["salt://grail/scene33"])
        with salt.utils.files.fopen(ret, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_get_url_no_dest(self):
        """
        cp.get_url with salt:// source given and destination set as None
        """
        tgt = None
        ret = self.run_function("cp.get_url", ["salt://grail/scene33", tgt])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)

    @pytest.mark.slow_test
    def test_get_url_nonexistent_source(self):
        """
        cp.get_url with nonexistent salt:// source given
        """
        tgt = None
        ret = self.run_function("cp.get_url", ["salt://grail/nonexistent_scene", tgt])
        self.assertEqual(ret, False)

    @pytest.mark.slow_test
    def test_get_url_to_dir(self):
        """
        cp.get_url with salt:// source
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "")
        self.run_function("cp.get_url", ["salt://grail/scene33", tgt])
        with salt.utils.files.fopen(tgt + "scene33", "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_url_https(self, tgt):
        """
        cp.get_url with https:// source given
        """
        self.run_function(
            "cp.get_url",
            ["https://packages.broadcom.com/artifactory/saltproject-generic/", tgt],
        )
        with salt.utils.files.fopen(tgt, "r") as instructions:
            data = salt.utils.stringutils.to_unicode(instructions.read())
        self.assertIn("Index of saltproject", data)
        self.assertIn("onedir", data)
        self.assertIn("Artifactory Online Server", data)
        self.assertNotIn("AYBABTU", data)

    @pytest.mark.slow_test
    def test_get_url_https_dest_empty(self):
        """
        cp.get_url with https:// source given and destination omitted.
        """
        ret = self.run_function(
            "cp.get_url",
            ["https://packages.broadcom.com/artifactory/saltproject-generic/"],
        )

        with salt.utils.files.fopen(ret, "r") as instructions:
            data = salt.utils.stringutils.to_unicode(instructions.read())
        self.assertIn("Index of saltproject", data)
        self.assertIn("onedir", data)
        self.assertIn("Artifactory Online Server", data)
        self.assertNotIn("AYBABTU", data)

    @pytest.mark.slow_test
    def test_get_url_https_no_dest(self):
        """
        cp.get_url with https:// source given and destination set as None
        """
        timeout = 500
        start = time.time()
        sleep = 5
        tgt = None
        while time.time() - start <= timeout:
            ret = self.run_function(
                "cp.get_url",
                ["https://packages.broadcom.com/artifactory/saltproject-generic/", tgt],
            )
            if ret.find("HTTP 599") == -1:
                break
            time.sleep(sleep)
        if ret.find("HTTP 599") != -1:
            raise Exception(
                "https://packages.broadcom.com/artifactory/saltproject-generic/ returned 599 error"
            )
        self.assertIn("Index of saltproject", ret)
        self.assertIn("onedir", ret)
        self.assertIn("Artifactory Online Server", ret)
        self.assertNotIn("AYBABTU", ret)

    @pytest.mark.slow_test
    def test_get_url_file(self):
        """
        cp.get_url with file:// source given
        """
        tgt = ""
        src = os.path.join("file://", RUNTIME_VARS.FILES, "file", "base", "file.big")
        ret = self.run_function("cp.get_url", [src, tgt])
        with salt.utils.files.fopen(ret, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_get_url_file_no_dest(self):
        """
        cp.get_url with file:// source given and destination set as None
        """
        tgt = None
        src = os.path.join("file://", RUNTIME_VARS.FILES, "file", "base", "file.big")
        ret = self.run_function("cp.get_url", [src, tgt])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)
        self.assertNotIn("bacon", ret)

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_url_ftp(self, tgt):
        """
        cp.get_url with https:// source given
        """
        self.run_function(
            "cp.get_url",
            [
                "ftp://ftp.freebsd.org/pub/FreeBSD/releases/amd64/README.TXT",
                tgt,
            ],
        )
        with salt.utils.files.fopen(tgt, "r") as instructions:
            data = salt.utils.stringutils.to_unicode(instructions.read())
        self.assertIn("The official FreeBSD", data)

    # cp.get_file_str tests

    @pytest.mark.slow_test
    def test_get_file_str_salt(self):
        """
        cp.get_file_str with salt:// source given
        """
        src = "salt://grail/scene33"
        ret = self.run_function("cp.get_file_str", [src])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)

    @pytest.mark.slow_test
    def test_get_file_str_nonexistent_source(self):
        """
        cp.get_file_str with nonexistent salt:// source given
        """
        src = "salt://grail/nonexistent_scene"
        ret = self.run_function("cp.get_file_str", [src])
        self.assertEqual(ret, False)

    @pytest.mark.slow_test
    def test_get_file_str_https(self):
        """
        cp.get_file_str with https:// source given
        """
        src = "https://packages.broadcom.com/artifactory/saltproject-generic/"
        ret = self.run_function("cp.get_file_str", [src])
        self.assertIn("Index of saltproject", ret)
        self.assertIn("onedir", ret)
        self.assertIn("Artifactory Online Server", ret)
        self.assertNotIn("AYBABTU", ret)

    @pytest.mark.slow_test
    def test_get_file_str_local(self):
        """
        cp.get_file_str with file:// source given
        """
        src = os.path.join("file://", RUNTIME_VARS.FILES, "file", "base", "file.big")
        ret = self.run_function("cp.get_file_str", [src])
        self.assertIn("KNIGHT:  They're nervous, sire.", ret)
        self.assertNotIn("bacon", ret)

    # caching tests

    @pytest.mark.slow_test
    def test_cache_file(self):
        """
        cp.cache_file
        """
        ret = self.run_function("cp.cache_file", ["salt://grail/scene33"])
        with salt.utils.files.fopen(ret, "r") as scene:
            data = salt.utils.stringutils.to_unicode(scene.read())
        self.assertIn("KNIGHT:  They're nervous, sire.", data)
        self.assertNotIn("bacon", data)

    @pytest.mark.slow_test
    def test_cache_files(self):
        """
        cp.cache_files
        """
        ret = self.run_function(
            "cp.cache_files", [["salt://grail/scene33", "salt://grail/36/scene"]]
        )
        for path in ret:
            with salt.utils.files.fopen(path, "r") as scene:
                data = salt.utils.stringutils.to_unicode(scene.read())
            self.assertIn("ARTHUR:", data)
            self.assertNotIn("bacon", data)

    @with_tempfile()
    @pytest.mark.slow_test
    def test_cache_master(self, tgt):
        """
        cp.cache_master
        """
        ret = self.run_function(
            "cp.cache_master",
            [tgt],
        )
        for path in ret:
            self.assertTrue(os.path.exists(path))

    @pytest.mark.slow_test
    def test_cache_local_file(self):
        """
        cp.cache_local_file
        """
        src = os.path.join(RUNTIME_VARS.TMP, "random")
        with salt.utils.files.fopen(src, "w+") as fn_:
            fn_.write(salt.utils.stringutils.to_str("foo"))
        ret = self.run_function("cp.cache_local_file", [src])
        with salt.utils.files.fopen(ret, "r") as cp_:
            self.assertEqual(salt.utils.stringutils.to_unicode(cp_.read()), "foo")

    @pytest.mark.skip_if_binaries_missing("nginx")
    @pytest.mark.slow_test
    @pytest.mark.skip_if_not_root
    def test_cache_remote_file(self):
        """
        cp.cache_file
        """
        nginx_port = ports.get_unused_localhost_port()
        url_prefix = f"http://localhost:{nginx_port}/"
        temp_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, temp_dir, ignore_errors=True)
        nginx_root_dir = os.path.join(temp_dir, "root")
        nginx_conf_dir = os.path.join(temp_dir, "conf")
        nginx_conf = os.path.join(nginx_conf_dir, "nginx.conf")
        nginx_pidfile = os.path.join(nginx_conf_dir, "nginx.pid")
        file_contents = "Hello world!"

        for dirname in (nginx_root_dir, nginx_conf_dir):
            os.makedirs(dirname)

        # Write the temp file
        with salt.utils.files.fopen(
            os.path.join(nginx_root_dir, "actual_file"), "w"
        ) as fp_:
            fp_.write(salt.utils.stringutils.to_str(file_contents))

        # Write the nginx config
        with salt.utils.files.fopen(nginx_conf, "w") as fp_:
            fp_.write(
                textwrap.dedent(
                    salt.utils.stringutils.to_str(
                        f"""\
                user root;
                worker_processes 1;
                error_log {nginx_conf_dir}/server_error.log;
                pid {nginx_pidfile};

                events {{
                    worker_connections 1024;
                }}

                http {{
                    include       /etc/nginx/mime.types;
                    default_type  application/octet-stream;

                    access_log {nginx_conf_dir}/access.log;
                    error_log {nginx_conf_dir}/error.log;

                    server {{
                        listen {nginx_port} default_server;
                        server_name cachefile.local;
                        root {nginx_root_dir};

                        location ~ ^/301$ {{
                            return 301 /actual_file;
                        }}

                        location ~ ^/302$ {{
                            return 302 /actual_file;
                        }}
                    }}
                }}"""
                    )
                )
            )

        self.run_function("cmd.run", [["nginx", "-c", nginx_conf]], python_shell=False)
        with salt.utils.files.fopen(nginx_pidfile) as fp_:
            nginx_pid = int(fp_.read().strip())
            nginx_proc = psutil.Process(pid=nginx_pid)
            self.addCleanup(nginx_proc.send_signal, signal.SIGQUIT)

        for code in ("", "301", "302"):
            url = url_prefix + (code or "actual_file")
            log.debug("attempting to cache %s", url)
            ret = self.run_function("cp.cache_file", [url])
            self.assertTrue(ret)
            with salt.utils.files.fopen(ret) as fp_:
                cached_contents = salt.utils.stringutils.to_unicode(fp_.read())
                self.assertEqual(cached_contents, file_contents)

    @pytest.mark.slow_test
    def test_list_states(self):
        """
        cp.list_states
        """
        top_sls = """
        base:
          '*':
            - core
            """

        core_state = """
        {}/testfile:
          file:
            - managed
            - source: salt://testfile
            - makedirs: true
            """.format(
            RUNTIME_VARS.TMP
        )

        with temp_file(
            "top.sls", top_sls, RUNTIME_VARS.TMP_BASEENV_STATE_TREE
        ), temp_file("core.sls", core_state, RUNTIME_VARS.TMP_BASEENV_STATE_TREE):
            ret = self.run_function(
                "cp.list_states",
            )
            self.assertIn("core", ret)
            self.assertIn("top", ret)

    @pytest.mark.slow_test
    def test_list_minion(self):
        """
        cp.list_minion
        """
        self.run_function("cp.cache_file", ["salt://grail/scene33"])
        ret = self.run_function("cp.list_minion")
        found = False
        search = "grail/scene33"
        if salt.utils.platform.is_windows():
            search = r"grail\scene33"
        for path in ret:
            if search in path:
                found = True
                break
        self.assertTrue(found)

    @pytest.mark.slow_test
    def test_is_cached(self):
        """
        cp.is_cached
        """
        self.run_function("cp.cache_file", ["salt://grail/scene33"])
        ret1 = self.run_function("cp.is_cached", ["salt://grail/scene33"])
        self.assertTrue(ret1)
        ret2 = self.run_function("cp.is_cached", ["salt://fasldkgj/poicxzbn"])
        self.assertFalse(ret2)

    @pytest.mark.slow_test
    def test_hash_file(self):
        """
        cp.hash_file
        """
        sha256_hash = self.run_function("cp.hash_file", ["salt://grail/scene33"])
        path = self.run_function("cp.cache_file", ["salt://grail/scene33"])
        with salt.utils.files.fopen(path, "rb") as fn_:
            data = fn_.read()
            self.assertEqual(sha256_hash["hsum"], hashlib.sha256(data).hexdigest())

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_file_from_env_predefined(self, tgt):
        """
        cp.get_file
        """
        tgt = os.path.join(RUNTIME_VARS.TMP, "cheese")
        try:
            self.run_function("cp.get_file", ["salt://cheese", tgt])
            with salt.utils.files.fopen(tgt, "r") as cheese:
                data = salt.utils.stringutils.to_unicode(cheese.read())
            self.assertIn("Gromit", data)
            self.assertNotIn("Comte", data)
        finally:
            os.unlink(tgt)

    @with_tempfile()
    @pytest.mark.slow_test
    def test_get_file_from_env_in_url(self, tgt):
        tgt = os.path.join(RUNTIME_VARS.TMP, "cheese")
        try:
            self.run_function("cp.get_file", ["salt://cheese?saltenv=prod", tgt])
            with salt.utils.files.fopen(tgt, "r") as cheese:
                data = salt.utils.stringutils.to_unicode(cheese.read())
            self.assertIn("Gromit", data)
            self.assertIn("Comte", data)
        finally:
            os.unlink(tgt)

    @pytest.mark.slow_test
    def test_push(self):
        log_to_xfer = os.path.join(RUNTIME_VARS.TMP, uuid.uuid4().hex)
        open(  # pylint: disable=resource-leakage
            log_to_xfer, "w", encoding="utf-8"
        ).close()
        try:
            self.run_function("cp.push", [log_to_xfer])
            tgt_cache_file = os.path.join(
                RUNTIME_VARS.RUNTIME_CONFIGS["master"]["cachedir"],
                "minions",
                "minion",
                "files",
                os.path.splitdrive(os.path.normpath(log_to_xfer.lstrip(os.sep)))[1],
            )
            self.assertTrue(
                os.path.isfile(tgt_cache_file), "File was not cached on the master"
            )
        finally:
            os.unlink(tgt_cache_file)

    @pytest.mark.slow_test
    def test_envs(self):
        self.assertEqual(sorted(self.run_function("cp.envs")), sorted(["base", "prod"]))



"""
    :codeauthor: :email:`Jakub Sliva <jakub.sliva@ultimum.io>`
"""

import pytest

import salt.states.zabbix_action as zabbix_action
from tests.support.mock import MagicMock, patch


@pytest.fixture
def input_params():
    return {
        "status": "0",
        "filter": {
            "evaltype": "2",
            "conditions": [
                {"operator": "2", "conditiontype": "24", "value": "database"}
            ],
        },
        "eventsource": "2",
        "name": "Auto registration Databases",
        "operations": [{"opgroup": [{"groupid": "6"}], "operationtype": "4"}],
    }


@pytest.fixture
def existing_obj():
    return [
        {
            "status": "0",
            "operations": [
                {
                    "operationtype": "4",
                    "esc_period": "0",
                    "evaltype": "0",
                    "opconditions": [],
                    "esc_step_to": "1",
                    "actionid": "28",
                    "esc_step_from": "1",
                    "opgroup": [{"groupid": "6", "operationid": "92"}],
                    "operationid": "92",
                }
            ],
            "def_shortdata": "",
            "name": "Auto registration Databases",
            "esc_period": "0",
            "def_longdata": "",
            "filter": {
                "formula": "",
                "evaltype": "2",
                "conditions": [
                    {
                        "operator": "2",
                        "conditiontype": "24",
                        "formulaid": "A",
                        "value": "database",
                    }
                ],
                "eval_formula": "A",
            },
            "eventsource": "2",
            "actionid": "28",
            "r_shortdata": "",
            "r_longdata": "",
            "recovery_msg": "0",
        }
    ]


@pytest.fixture
def existing_obj_diff():
    return {
        "status": "0",
        "operations": [
            {
                "operationtype": "4",
                "esc_period": "0",
                "evaltype": "0",
                "opconditions": [],
                "esc_step_to": "1",
                "actionid": "28",
                "esc_step_from": "1",
                "opgroup": [{"groupid": "6", "operationid": "92"}],
                "operationid": "92",
            }
        ],
        "def_shortdata": "",
        "name": "Auto registration Databases",
        "esc_period": "0",
        "def_longdata": "",
        "filter": {
            "formula": "",
            "evaltype": "2",
            "conditions": [
                {
                    "operator": "2",
                    "conditiontype": "24",
                    "formulaid": "A",
                    "value": "SOME OTHER VALUE",
                }
            ],
            "eval_formula": "A",
        },
        "eventsource": "2",
        "actionid": "28",
        "r_shortdata": "",
        "r_longdata": "",
        "recovery_msg": "0",
    }


@pytest.fixture
def diff_params():
    return {
        "filter": {
            "evaltype": "2",
            "conditions": [
                {"operator": "2", "conditiontype": "24", "value": "virtual"}
            ],
        },
        "actionid": "28",
    }


@pytest.fixture
def configure_loader_modules():
    return {zabbix_action: {}}


def test_present_create(input_params):
    """
    Test to ensure that named action is created
    """
    name = "Auto registration Databases"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    def side_effect_run_query(*args):
        """
        Differentiate between __salt__ exec module function calls with different parameters.
        """
        if args[0] == "action.get":
            return False
        elif args[0] == "action.create":
            return True

    with patch.dict(zabbix_action.__opts__, {"test": False}):
        with patch.dict(
            zabbix_action.__salt__,
            {
                "zabbix.get_zabbix_id_mapper": MagicMock(
                    return_value={"action": "actionid"}
                ),
                "zabbix.substitute_params": MagicMock(
                    side_effect=[input_params, False]
                ),
                "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                "zabbix.compare_params": MagicMock(return_value={}),
            },
        ):
            ret["result"] = True
            ret["comment"] = f'Zabbix Action "{name}" created.'
            ret["changes"] = {
                name: {
                    "old": f'Zabbix Action "{name}" did not exist.',
                    "new": 'Zabbix Action "{}" created according definition.'.format(
                        name
                    ),
                }
            }
            assert zabbix_action.present(name, {}) == ret


def test_present_exists(input_params, existing_obj):
    """
    Test to ensure that named action is present and not changed
    """
    name = "Auto registration Databases"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    with patch.dict(zabbix_action.__opts__, {"test": False}):
        with patch.dict(
            zabbix_action.__salt__,
            {
                "zabbix.get_zabbix_id_mapper": MagicMock(
                    return_value={"action": "actionid"}
                ),
                "zabbix.substitute_params": MagicMock(
                    side_effect=[input_params, existing_obj]
                ),
                "zabbix.run_query": MagicMock(return_value=["length of result is 1"]),
                "zabbix.compare_params": MagicMock(return_value={}),
            },
        ):
            ret["result"] = True
            ret["comment"] = (
                'Zabbix Action "{}" already exists and corresponds to a definition.'.format(
                    name
                )
            )
            assert zabbix_action.present(name, {}) == ret


def test_present_update(input_params, existing_obj_diff, diff_params):
    """
    Test to ensure that named action is present but must be updated
    """
    name = "Auto registration Databases"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    def side_effect_run_query(*args):
        """
        Differentiate between __salt__ exec module function calls with different parameters.
        """
        if args[0] == "action.get":
            return ["length of result is 1 = action exists"]
        elif args[0] == "action.update":
            return diff_params

    with patch.dict(zabbix_action.__opts__, {"test": False}):
        with patch.dict(
            zabbix_action.__salt__,
            {
                "zabbix.get_zabbix_id_mapper": MagicMock(
                    return_value={"action": "actionid"}
                ),
                "zabbix.substitute_params": MagicMock(
                    side_effect=[input_params, existing_obj_diff]
                ),
                "zabbix.run_query": MagicMock(side_effect=side_effect_run_query),
                "zabbix.compare_params": MagicMock(return_value=diff_params),
            },
        ):
            ret["result"] = True
            ret["comment"] = f'Zabbix Action "{name}" updated.'
            ret["changes"] = {
                name: {
                    "old": (
                        'Zabbix Action "{}" differed '
                        "in following parameters: {}".format(name, diff_params)
                    ),
                    "new": f'Zabbix Action "{name}" fixed.',
                }
            }
            assert zabbix_action.present(name, {}) == ret


def test_absent_test_mode():
    """
    Test to ensure that named action is absent in test mode
    """
    name = "Auto registration Databases"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    with patch.dict(zabbix_action.__opts__, {"test": True}):
        with patch.dict(
            zabbix_action.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
        ):
            ret["result"] = True
            ret["comment"] = f'Zabbix Action "{name}" would be deleted.'
            ret["changes"] = {
                name: {
                    "old": f'Zabbix Action "{name}" exists.',
                    "new": f'Zabbix Action "{name}" would be deleted.',
                }
            }
            assert zabbix_action.absent(name) == ret


def test_absent():
    """
    Test to ensure that named action is absent
    """
    name = "Auto registration Databases"
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    with patch.dict(zabbix_action.__opts__, {"test": False}):
        with patch.dict(
            zabbix_action.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=False)},
        ):
            ret["result"] = True
            ret["comment"] = f'Zabbix Action "{name}" does not exist.'
            assert zabbix_action.absent(name) == ret

        with patch.dict(
            zabbix_action.__salt__,
            {"zabbix.get_object_id_by_params": MagicMock(return_value=11)},
        ):
            with patch.dict(
                zabbix_action.__salt__,
                {"zabbix.run_query": MagicMock(return_value=True)},
            ):
                ret["result"] = True
                ret["comment"] = f'Zabbix Action "{name}" deleted.'
                ret["changes"] = {
                    name: {
                        "old": f'Zabbix Action "{name}" existed.',
                        "new": f'Zabbix Action "{name}" deleted.',
                    }
                }
                assert zabbix_action.absent(name) == ret



import os.path
import pathlib
import subprocess
import time

import pytest
from pytestskipmarkers.utils import platform


@pytest.mark.skip_on_windows
def test_salt_version(version, install_salt):
    """
    Test version output from salt --version
    """
    actual = []
    test_bin = os.path.join(*install_salt.binary_paths["salt"])
    ret = install_salt.proc.run(test_bin, "--version")
    if "+" in version:
        # testing a non-release build artifact version
        actual = ret.stdout.strip().split(" ")[:2]
    else:
        # testing against release build version, for example: downgrade
        actual_ver = ret.stdout.strip().split(" ")[:2]
        actual_ver_salt = actual_ver[1]  # get salt version
        if "+" in actual_ver_salt:
            actual_ver_salt_stripped = actual_ver_salt.split("+")[
                0
            ]  # strip any git versioning
            actual.append(actual_ver[0])
            actual.append(actual_ver_salt_stripped)
        else:
            pytest.skip("We don't run this test on release builds")

    expected = ["salt", version]
    assert actual == expected


@pytest.mark.skip_on_windows
@pytest.mark.skip_on_darwin
def test_salt_versions_report_master(install_salt):
    """
    Test running --versions-report on master
    """
    test_bin = os.path.join(*install_salt.binary_paths["master"])
    python_bin = os.path.join(*install_salt.binary_paths["python"])
    ret = install_salt.proc.run(test_bin, "--versions-report")
    ret.stdout.matcher.fnmatch_lines(["*Salt Version:*"])
    py_version = subprocess.run(
        [str(python_bin), "--version"],
        check=True,
        capture_output=True,
    ).stdout
    py_version = py_version.decode().strip().replace(" ", ": ")
    ret.stdout.matcher.fnmatch_lines([f"*{py_version}*"])


@pytest.mark.skip_on_windows
def test_salt_versions_report_minion(salt_cli, salt_call_cli, salt_master, salt_minion):
    """
    Test running test.versions_report on minion
    """
    # Make sure the minion is running
    for count in range(0, 30):
        if salt_minion.is_running():
            break
        else:
            time.sleep(2)

    assert salt_minion.is_running()

    # Make sure the master is running
    for count in range(0, 30):
        if salt_master.is_running():
            break
        else:
            time.sleep(2)

    assert salt_master.is_running()

    # Make sure we can ping the minion ...
    ret = salt_cli.run(
        "--timeout=600", "test.ping", minion_tgt=salt_minion.id, _timeout=600
    )

    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_cli.run(
        "--hard-crash",
        "--failhard",
        "--timeout=300",
        "test.versions_report",
        minion_tgt=salt_minion.id,
        _timeout=300,
    )
    ret.stdout.matcher.fnmatch_lines(["*Salt Version:*"])


@pytest.mark.skip_on_windows
@pytest.mark.skip_on_darwin
@pytest.mark.parametrize(
    "binary", ["master", "cloud", "syndic", "minion", "call", "api"]
)
def test_compare_versions(binary, install_salt):
    """
    Test compare versions
    """
    if install_salt.use_prev_version:
        version = install_salt.prev_version
    else:
        version = install_salt.artifact_version
    if binary in install_salt.binary_paths:
        if install_salt.upgrade:
            install_salt.install()

        ret = install_salt.proc.run(
            *install_salt.binary_paths[binary],
            "--version",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        ret.stdout.matcher.fnmatch_lines([f"*{version}*"])
    else:
        if platform.is_windows():
            pytest.skip(f"Binary not available on windows: {binary}")
        pytest.fail(
            f"Platform is not Windows and yet the binary {binary!r} is not available"
        )


@pytest.mark.skip_unless_on_darwin
@pytest.mark.parametrize(
    "symlink",
    [
        "salt",
        "salt-api",
        "salt-call",
        "salt-cloud",
        "salt-cp",
        "salt-key",
        "salt-master",
        "salt-minion",
        "salt-proxy",
        "salt-run",
        "spm",
        "salt-ssh",
        "salt-syndic",
    ],
)
def test_symlinks_created(version, symlink, install_salt):
    """
    Test symlinks created
    """
    ret = install_salt.proc.run(pathlib.Path("/usr/local/sbin") / symlink, "--version")
    install_log_file = pathlib.Path("/tmp") / "postinstall.txt"
    install_log_content = install_log_file.read_text()
    ret.stdout.matcher.fnmatch_lines([f"*{version}*"])


@pytest.mark.skip_unless_on_linux
def test_compare_pkg_versions_redhat_rc(version, install_salt):
    """
    Test compare pkg versions for redhat RC packages. A tilde should be included
    in RC Packages and it should test to be a lower version than a non RC
    package of the same version. For example, v3004~rc1 should be less than
    v3004.
    """
    if install_salt.distro_id not in (
        "almalinux",
        "rocky",
        "centos",
        "redhat",
        "amzn",
        "fedora",
        "photon",
    ):
        pytest.skip("Only tests rpm packages")

    pkg = [x for x in install_salt.pkgs if "rpm" in x]
    if not pkg:
        pytest.skip("Not testing rpm packages")
    pkg = pkg[0].split("/")[-1]
    if "rc" not in ".".join(pkg.split(".")[:2]):
        pytest.skip("Not testing an RC package")
    assert "~" in pkg
    comp_pkg = pkg.split("~")[0]
    ret = install_salt.proc.run("rpmdev-vercmp", pkg, comp_pkg)
    ret.stdout.matcher.fnmatch_lines([f"{pkg} < {comp_pkg}"])



"""
The default file server backend

This fileserver backend serves files from the Master's local filesystem. If
:conf_master:`fileserver_backend` is not defined in the Master config file,
then this backend is enabled by default. If it *is* defined then ``roots`` must
be in the :conf_master:`fileserver_backend` list to enable this backend.

.. code-block:: yaml

    fileserver_backend:
      - roots

Fileserver environments are defined using the :conf_master:`file_roots`
configuration option.
"""

import errno
import logging
import os

import salt.fileserver
import salt.utils.event
import salt.utils.files
import salt.utils.gzip_util
import salt.utils.hashutils
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.verify
import salt.utils.versions

log = logging.getLogger(__name__)


def find_file(path, saltenv="base", **kwargs):
    """
    Search the environment for the relative path.
    """
    actual_saltenv = saltenv
    if "env" in kwargs:
        # "env" is not supported; Use "saltenv".
        kwargs.pop("env")

    path = os.path.normpath(path)
    fnd = {"path": "", "rel": ""}
    if os.path.isabs(path):
        return fnd
    if saltenv not in __opts__["file_roots"]:
        if "__env__" in __opts__["file_roots"]:
            log.debug(
                "salt environment '%s' maps to __env__ file_roots directory", saltenv
            )
            saltenv = "__env__"
        else:
            return fnd

    def _add_file_stat(fnd):
        """
        Stat the file and, assuming no errors were found, convert the stat
        result to a list of values and add to the return dict.

        Converting the stat result to a list, the elements of the list
        correspond to the following stat_result params:

        0 => st_mode=33188
        1 => st_ino=10227377
        2 => st_dev=65026
        3 => st_nlink=1
        4 => st_uid=1000
        5 => st_gid=1000
        6 => st_size=1056233
        7 => st_atime=1468284229
        8 => st_mtime=1456338235
        9 => st_ctime=1456338235
        """
        try:
            fnd["stat"] = list(os.stat(fnd["path"]))
        except Exception as exc:  # pylint: disable=broad-except
            log.error("Unable to stat file: %s", exc)
        return fnd

    if "index" in kwargs:
        try:
            root = __opts__["file_roots"][saltenv][int(kwargs["index"])]
        except IndexError:
            # An invalid index was passed
            return fnd
        except ValueError:
            # An invalid index option was passed
            return fnd
        full = os.path.join(root, path)
        if os.path.isfile(full) and not salt.fileserver.is_file_ignored(__opts__, full):
            fnd["path"] = full
            fnd["rel"] = path
            return _add_file_stat(fnd)
        return fnd
    for root in __opts__["file_roots"][saltenv]:
        if saltenv == "__env__":
            root = root.replace("__env__", actual_saltenv)
        full = os.path.join(root, path)

        # Refuse to serve file that is not under the root.
        if not salt.utils.verify.clean_path(
            root, full, subdir=True, realpath=not __opts__["fileserver_followsymlinks"]
        ):
            continue

        if os.path.isfile(full) and not salt.fileserver.is_file_ignored(__opts__, full):
            fnd["path"] = full
            fnd["rel"] = path
            return _add_file_stat(fnd)
    return fnd


def envs():
    """
    Return the file server environments
    """
    return sorted(__opts__["file_roots"])


def serve_file(load, fnd):
    """
    Return a chunk from a file based on the data received
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    ret = {"data": "", "dest": ""}
    if "path" not in load or "loc" not in load or "saltenv" not in load:
        return ret
    if not fnd["path"]:
        return ret
    ret["dest"] = fnd["rel"]
    gzip = load.get("gzip", None)
    fpath = os.path.normpath(fnd["path"])

    actual_saltenv = saltenv = load["saltenv"]
    if saltenv not in __opts__["file_roots"]:
        if "__env__" in __opts__["file_roots"]:
            log.debug(
                "salt environment '%s' maps to __env__ file_roots directory", saltenv
            )
            saltenv = "__env__"
        else:
            return fnd
    file_in_root = False
    for root in __opts__["file_roots"][saltenv]:
        if saltenv == "__env__":
            root = root.replace("__env__", actual_saltenv)
        # Refuse to serve file that is not under the root.
        if salt.utils.verify.clean_path(
            root, fpath, subdir=True, realpath=not __opts__["fileserver_followsymlinks"]
        ):
            file_in_root = True
    if not file_in_root:
        return ret

    with salt.utils.files.fopen(fpath, "rb") as fp_:
        fp_.seek(load["loc"])
        data = fp_.read(__opts__["file_buffer_size"])
        if gzip and data:
            data = salt.utils.gzip_util.compress(data, gzip)
            ret["gzip"] = gzip
        ret["data"] = data
    return ret


def update():
    """
    When we are asked to update (regular interval) lets reap the cache
    """
    try:
        salt.fileserver.reap_fileserver_cache_dir(
            os.path.join(__opts__["cachedir"], "roots", "hash"), find_file
        )
    except OSError:
        # Hash file won't exist if no files have yet been served up
        pass

    mtime_map_path = os.path.join(__opts__["cachedir"], "roots", "mtime_map")
    # data to send on event
    data = {"changed": False, "files": {"changed": []}, "backend": "roots"}

    # generate the new map
    new_mtime_map = salt.fileserver.generate_mtime_map(__opts__, __opts__["file_roots"])

    old_mtime_map = {}
    # if you have an old map, load that
    try:
        with salt.utils.files.fopen(mtime_map_path, encoding="utf-8") as fp_:
            for line in fp_:
                try:
                    file_path, mtime = line.strip().rsplit(":", 1)
                    mtime = float(mtime)
                    old_mtime_map[file_path] = mtime
                    if mtime != new_mtime_map.get(file_path, mtime):
                        data["files"]["changed"].append(file_path)
                except ValueError:
                    # Document the invalid entry in the log
                    log.warning(
                        "Skipped invalid cache mtime entry in %s: %s",
                        mtime_map_path,
                        line,
                    )
    except (OSError, UnicodeDecodeError):
        pass

    # compare the maps, set changed to the return value
    data["changed"] = salt.fileserver.diff_mtime_map(old_mtime_map, new_mtime_map)

    # compute files that were removed and added
    old_files = set(old_mtime_map)
    new_files = set(new_mtime_map)
    data["files"]["removed"] = list(old_files - new_files)
    data["files"]["added"] = list(new_files - old_files)

    # write out the new map
    mtime_map_path_dir = os.path.dirname(mtime_map_path)
    if not os.path.exists(mtime_map_path_dir):
        os.makedirs(mtime_map_path_dir)
    with salt.utils.files.fopen(mtime_map_path, "wb") as fp_:
        for file_path, mtime in new_mtime_map.items():
            fp_.write(salt.utils.stringutils.to_bytes(f"{file_path}:{mtime}\n"))

    if __opts__.get("fileserver_events", False):
        # if there is a change, fire an event
        with salt.utils.event.get_event(
            "master",
            __opts__["sock_dir"],
            opts=__opts__,
            listen=False,
        ) as event:
            event.fire_event(
                data, salt.utils.event.tagify(["roots", "update"], prefix="fileserver")
            )
    # return data is used for tests
    # but can also be used to get file changes with out needing fileserver events
    return data


def file_hash(load, fnd):
    """
    Return a file hash, the hash type is set in the master config file
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    if "path" not in load or "saltenv" not in load:
        return ""
    path = fnd["path"]
    saltenv = load["saltenv"]
    if saltenv not in __opts__["file_roots"] and "__env__" in __opts__["file_roots"]:
        saltenv = "__env__"
    ret = {}

    # if the file doesn't exist, we can't get a hash
    if not path or not os.path.isfile(path):
        return ret

    # set the hash_type as it is determined by config-- so mechanism won't change that
    ret["hash_type"] = __opts__["hash_type"]

    # check if the hash is cached
    # cache file's contents should be "hash:mtime"
    cache_path = os.path.join(
        __opts__["cachedir"],
        "roots",
        "hash",
        saltenv,
        "{}.hash.{}".format(fnd["rel"], __opts__["hash_type"]),
    )
    # if we have a cache, serve that if the mtime hasn't changed
    if os.path.exists(cache_path):
        try:
            with salt.utils.files.fopen(cache_path, encoding="utf-8") as fp_:
                try:
                    hsum, mtime = fp_.read().split(":")
                except ValueError:
                    log.debug(
                        "Fileserver attempted to read incomplete cache file. Retrying."
                    )
                    # Delete the file since its incomplete (either corrupted or incomplete)
                    try:
                        os.unlink(cache_path)
                    except OSError:
                        pass
                    return file_hash(load, fnd)
                if str(os.path.getmtime(path)) == mtime:
                    # check if mtime changed
                    ret["hsum"] = hsum
                    return ret
        except OSError:  # Can't use Python select() because we need Windows support
            log.debug("Fileserver encountered lock when reading cache file. Retrying.")
            # Delete the file since its incomplete (either corrupted or incomplete)
            try:
                os.unlink(cache_path)
            except OSError:
                pass
            return file_hash(load, fnd)

    # if we don't have a cache entry-- lets make one
    ret["hsum"] = salt.utils.hashutils.get_hash(path, __opts__["hash_type"])
    cache_dir = os.path.dirname(cache_path)
    # make cache directory if it doesn't exist
    if not os.path.exists(cache_dir):
        try:
            os.makedirs(cache_dir)
        except OSError as err:
            if err.errno == errno.EEXIST:
                # rarely, the directory can be already concurrently created between
                # the os.path.exists and the os.makedirs lines above
                pass
            else:
                raise
    # save the cache object "hash:mtime"
    cache_object = "{}:{}".format(ret["hsum"], os.path.getmtime(path))
    with salt.utils.files.flopen(cache_path, "w") as fp_:
        fp_.write(cache_object)
    return ret


def _file_lists(load, form):
    """
    Return a dict containing the file lists for files, dirs, empty dirs and symlinks
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    saltenv = load["saltenv"]
    actual_saltenv = saltenv
    if saltenv not in __opts__["file_roots"]:
        if "__env__" in __opts__["file_roots"]:
            log.debug(
                "salt environment '%s' maps to __env__ file_roots directory", saltenv
            )
            saltenv = "__env__"
        else:
            return []

    list_cachedir = os.path.join(__opts__["cachedir"], "file_lists", "roots")
    if not os.path.isdir(list_cachedir):
        try:
            os.makedirs(list_cachedir)
        except OSError:
            log.critical("Unable to make cachedir %s", list_cachedir)
            return []
    list_cache = os.path.join(
        list_cachedir,
        f"{salt.utils.files.safe_filename_leaf(actual_saltenv)}.p",
    )
    w_lock = os.path.join(
        list_cachedir,
        f".{salt.utils.files.safe_filename_leaf(actual_saltenv)}.w",
    )
    cache_match, refresh_cache, save_cache = salt.fileserver.check_file_list_cache(
        __opts__, form, list_cache, w_lock
    )
    if cache_match is not None:
        return cache_match
    if refresh_cache:
        ret = {"files": set(), "dirs": set(), "empty_dirs": set(), "links": {}}

        def _add_to(tgt, fs_root, parent_dir, items):
            """
            Add the files to the target set
            """

            def _translate_sep(path):
                """
                Translate path separators for Windows masterless minions
                """
                return path.replace("\\", "/") if os.path.sep == "\\" else path

            for item in items:
                abs_path = os.path.join(parent_dir, item)
                log.trace("roots: Processing %s", abs_path)
                is_link = salt.utils.path.islink(abs_path)
                log.trace(
                    "roots: %s is %sa link", abs_path, "not " if not is_link else ""
                )
                if is_link and __opts__["fileserver_ignoresymlinks"]:
                    continue
                rel_path = _translate_sep(os.path.relpath(abs_path, fs_root))
                log.trace("roots: %s relative path is %s", abs_path, rel_path)
                if salt.fileserver.is_file_ignored(__opts__, rel_path):
                    continue
                tgt.add(rel_path)
                if os.path.isdir(abs_path):
                    try:
                        if not os.listdir(abs_path):
                            ret["empty_dirs"].add(rel_path)
                    except OSError:
                        log.debug("Unable to list dir: %s", abs_path)
                if is_link:
                    link_dest = salt.utils.path.readlink(abs_path)
                    log.trace(
                        "roots: %s symlink destination is %s", abs_path, link_dest
                    )
                    if salt.utils.platform.is_windows() and link_dest.startswith(
                        "\\\\"
                    ):
                        # Symlink points to a network path. Since you can't
                        # join UNC and non-UNC paths, just assume the original
                        # path.
                        log.trace(
                            "roots: %s is a UNC path, using %s instead",
                            link_dest,
                            abs_path,
                        )
                        link_dest = abs_path
                    if link_dest.startswith(".."):
                        joined = os.path.join(abs_path, link_dest)
                    else:
                        joined = os.path.join(os.path.dirname(abs_path), link_dest)
                    rel_dest = _translate_sep(
                        os.path.relpath(
                            os.path.realpath(os.path.normpath(joined)),
                            os.path.realpath(fs_root),
                        )
                    )
                    log.trace("roots: %s relative path is %s", abs_path, rel_dest)
                    if not rel_dest.startswith(".."):
                        # Only count the link if it does not point
                        # outside of the root dir of the fileserver
                        # (i.e. the "path" variable)
                        ret["links"][rel_path] = link_dest
                    else:
                        if not __opts__["fileserver_followsymlinks"]:
                            ret["links"][rel_path] = link_dest

        for path in __opts__["file_roots"][saltenv]:
            if saltenv == "__env__":
                path = path.replace("__env__", actual_saltenv)
            for root, dirs, files in salt.utils.path.os_walk(
                path, followlinks=__opts__["fileserver_followsymlinks"]
            ):
                _add_to(ret["dirs"], path, root, dirs)
                _add_to(ret["files"], path, root, files)

        ret["files"] = sorted(ret["files"])
        ret["dirs"] = sorted(ret["dirs"])
        ret["empty_dirs"] = sorted(ret["empty_dirs"])

        if save_cache:
            try:
                salt.fileserver.write_file_list_cache(__opts__, ret, list_cache, w_lock)
            except NameError:
                # Catch msgpack error in salt-ssh
                pass
        return ret.get(form, [])
    # Shouldn't get here, but if we do, this prevents a TypeError
    return []


def file_list(load):
    """
    Return a list of all files on the file server in a specified
    environment
    """
    return _file_lists(load, "files")


def file_list_emptydirs(load):
    """
    Return a list of all empty directories on the master
    """
    return _file_lists(load, "empty_dirs")


def dir_list(load):
    """
    Return a list of all directories on the master
    """
    return _file_lists(load, "dirs")


def symlink_list(load):
    """
    Return a dict of all symlinks based on a given path on the Master
    """
    if "env" in load:
        # "env" is not supported; Use "saltenv".
        load.pop("env")

    ret = {}
    if (
        load["saltenv"] not in __opts__["file_roots"]
        and "__env__" not in __opts__["file_roots"]
    ):
        return ret

    if "prefix" in load:
        prefix = load["prefix"].strip("/")
    else:
        prefix = ""

    symlinks = _file_lists(load, "links")
    return {key: val for key, val in symlinks.items() if key.startswith(prefix)}



"""
Test cases for salt.modules.openvswitch.
"""

import pytest

import salt.modules.openvswitch as openvswitch
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {openvswitch: {}}


def test_bridge_create_may_not_exist():
    """
    Test bridge_create function.

    This tests the case where neither a parent nor the may-exists flag are
    specified.
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        ret = openvswitch.bridge_create("br0", False)
        assert ret is True
        mock.assert_called_with("ovs-vsctl add-br br0")


def test_bridge_create_may_exist():
    """
    Test bridge_create function.

    This tests the case where no parent but the may-exists flag is specified.
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        ret = openvswitch.bridge_create("br1", True)
        assert ret is True
        mock.assert_called_with("ovs-vsctl --may-exist add-br br1")


def test_bridge_create_with_parent_may_exist():
    """
    Test bridge_create function.

    This tests the case where a parent is specified but the may-exists flag is
    false.
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        ret = openvswitch.bridge_create("br2", False, "br0", 42)
        assert ret is True
        mock.assert_called_with("ovs-vsctl add-br br2 br0 42")


def test_bridge_to_parent():
    """
    Test bridge_to_parent function.
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": "br0\n"})
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        ret = openvswitch.bridge_to_parent("br1")
        assert ret == "br0"
        mock.assert_called_with("ovs-vsctl br-to-parent br1")


def test_bridge_to_vlan():
    """
    Test bridge_to_vlan function.
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": "42\n"})
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        ret = openvswitch.bridge_to_vlan("br0")
        assert ret == 42
        mock.assert_called_with("ovs-vsctl br-to-vlan br0")


def test_db_get():
    """
    Test db_get function.
    """
    mock = MagicMock(
        return_value={
            "retcode": 0,
            "stdout": '{"data":[["01:02:03:04:05:06"]],' '"headings":["mac"]}',
        }
    )
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        ret = openvswitch.db_get("Interface", "br0", "mac")
        assert ret == "01:02:03:04:05:06"
        mock.assert_called_with(
            [
                "ovs-vsctl",
                "--format=json",
                "--columns=mac",
                "list",
                "Interface",
                "br0",
            ]
        )


def test_db_set():
    """
    Test db_set function.
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(
        openvswitch.__salt__,
        {"cmd.run_all": mock},
    ):
        openvswitch.db_set("Interface", "br0", "mac", "01:02:03:04:05:06")
        mock.assert_called_with(
            ["ovs-vsctl", "set", "Interface", "br0", 'mac="01:02:03:04:05:06"']
        )
